# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from unittest.mock import Mock

from twisted.internet import defer
from twisted.trial import unittest

from buildbot import config
from buildbot.process import results
from buildbot.process.buildstep import create_step_from_step_or_factory
from buildbot.process.properties import Property
from buildbot.process.results import FAILURE
from buildbot.process.results import SKIPPED
from buildbot.process.results import SUCCESS
from buildbot.process.results import WARNINGS
from buildbot.steps import vstudio
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.steps import ExpectShell
from buildbot.test.steps import TestBuildStepMixin

real_log = r"""
1>------ Build started: Project: lib1, Configuration: debug Win32 ------
1>Compiling...
1>SystemLog.cpp
1>c:\absolute\path\to\systemlog.cpp(7) : warning C4100: 'op' : unreferenced formal parameter
1>c:\absolute\path\to\systemlog.cpp(12) : warning C4100: 'statusword' : unreferenced formal parameter
1>c:\absolute\path\to\systemlog.cpp(12) : warning C4100: 'op' : unreferenced formal parameter
1>c:\absolute\path\to\systemlog.cpp(17) : warning C4100: 'retryCounter' : unreferenced formal parameter
1>c:\absolute\path\to\systemlog.cpp(17) : warning C4100: 'op' : unreferenced formal parameter
1>c:\absolute\path\to\systemlog.cpp(22) : warning C4100: 'op' : unreferenced formal parameter
1>Creating library...
1>Build log was saved at "file://c:\another\absolute\path\to\debug\BuildLog.htm"
1>lib1 - 0 error(s), 6 warning(s)
2>------ Build started: Project: product, Configuration: debug Win32 ------
2>Linking...
2>LINK : fatal error LNK1168: cannot open ../../debug/directory/dllname.dll for writing
2>Build log was saved at "file://c:\another\similar\path\to\debug\BuildLog.htm"
2>product - 1 error(s), 0 warning(s)
========== Build: 1 succeeded, 1 failed, 6 up-to-date, 0 skipped ==========
"""


class TestAddEnvPath(unittest.TestCase):
    def do_test(self, initial_env, name, value, expected_env):
        step = create_step_from_step_or_factory(vstudio.VisualStudio())
        step.env = initial_env
        step.add_env_path(name, value)
        self.assertEqual(step.env, expected_env)

    def test_new(self):
        self.do_test({}, 'PATH', r'C:\NOTHING', {'PATH': r'C:\NOTHING;'})

    def test_new_semi(self):
        self.do_test({}, 'PATH', r'C:\NOTHING;', {'PATH': r'C:\NOTHING;'})

    def test_existing(self):
        self.do_test({'PATH': '/bin'}, 'PATH', r'C:\NOTHING', {'PATH': r'/bin;C:\NOTHING;'})

    def test_existing_semi(self):
        self.do_test({'PATH': '/bin;'}, 'PATH', r'C:\NOTHING', {'PATH': r'/bin;C:\NOTHING;'})

    def test_existing_both_semi(self):
        self.do_test({'PATH': '/bin;'}, 'PATH', r'C:\NOTHING;', {'PATH': r'/bin;C:\NOTHING;'})


class MSLogLineObserver(unittest.TestCase):
    def setUp(self):
        self.warnings = []
        lw = Mock()
        lw.addStdout = lambda l: self.warnings.append(l.rstrip())

        self.errors = []
        self.errors_stderr = []
        le = Mock()
        le.addStdout = lambda l: self.errors.append(('o', l.rstrip()))
        le.addStderr = lambda l: self.errors.append(('e', l.rstrip()))

        self.llo = vstudio.MSLogLineObserver(lw, le)

        self.progress = {}
        self.llo.step = Mock()
        self.llo.step.setProgress = self.progress.__setitem__

    def receiveLines(self, *lines):
        for line in lines:
            self.llo.outLineReceived(line)

    def assertResult(
        self,
        nbFiles=0,
        nbProjects=0,
        nbWarnings=0,
        nbErrors=0,
        errors=None,
        warnings=None,
        progress=None,
    ):
        if errors is None:
            errors = []
        if warnings is None:
            warnings = []
        if progress is None:
            progress = {}
        self.assertEqual(
            {
                "nbFiles": self.llo.nbFiles,
                "nbProjects": self.llo.nbProjects,
                "nbWarnings": self.llo.nbWarnings,
                "nbErrors": self.llo.nbErrors,
                "errors": self.errors,
                "warnings": self.warnings,
                "progress": self.progress,
            },
            {
                "nbFiles": nbFiles,
                "nbProjects": nbProjects,
                "nbWarnings": nbWarnings,
                "nbErrors": nbErrors,
                "errors": errors,
                "warnings": warnings,
                "progress": progress,
            },
        )

    def test_outLineReceived_empty(self):
        self.llo.outLineReceived('abcd\r\n')
        self.assertResult()

    def test_outLineReceived_projects(self):
        lines = [
            "123>----- some project 1 -----",
            "123>----- some project 2 -----",
        ]
        self.receiveLines(*lines)
        self.assertResult(
            nbProjects=2, progress={"projects": 2}, errors=[('o', l) for l in lines], warnings=lines
        )

    def test_outLineReceived_files(self):
        lines = [
            "123>SomeClass.cpp",
            "123>SomeStuff.c",
            "123>SomeStuff.h",  # .h files not recognized
        ]
        self.receiveLines(*lines)
        self.assertResult(nbFiles=2, progress={"files": 2})

    def test_outLineReceived_warnings(self):
        lines = [
            "abc: warning ABC123: xyz!",
            "def : warning DEF456: wxy!",
        ]
        self.receiveLines(*lines)
        self.assertResult(nbWarnings=2, progress={"warnings": 2}, warnings=lines)

    def test_outLineReceived_errors(self):
        lines = [
            "error ABC123: foo",
            " error DEF456 : bar",
            " error : bar",
            " error: bar",  # NOTE: not matched
        ]
        self.receiveLines(*lines)
        self.assertResult(
            nbErrors=3,  # note: no progress
            errors=[
                ('e', "error ABC123: foo"),
                ('e', " error DEF456 : bar"),
                ('e', " error : bar"),
            ],
        )

    def test_outLineReceived_real(self):
        # based on a real logfile donated by Ben Allard
        lines = real_log.split("\n")
        self.receiveLines(*lines)
        errors = [
            ('o', '1>------ Build started: Project: lib1, Configuration: debug Win32 ------'),
            ('o', '2>------ Build started: Project: product, Configuration: debug Win32 ------'),
            (
                'e',
                '2>LINK : fatal error LNK1168: cannot open ../../debug/directory/dllname.dll for writing',
            ),
        ]
        warnings = [
            '1>------ Build started: Project: lib1, Configuration: debug Win32 ------',
            "1>c:\\absolute\\path\\to\\systemlog.cpp(7) : warning C4100: 'op' : unreferenced formal parameter",
            "1>c:\\absolute\\path\\to\\systemlog.cpp(12) : warning C4100: 'statusword' : unreferenced formal parameter",
            "1>c:\\absolute\\path\\to\\systemlog.cpp(12) : warning C4100: 'op' : unreferenced formal parameter",
            "1>c:\\absolute\\path\\to\\systemlog.cpp(17) : warning C4100: 'retryCounter' : unreferenced formal parameter",
            "1>c:\\absolute\\path\\to\\systemlog.cpp(17) : warning C4100: 'op' : unreferenced formal parameter",
            "1>c:\\absolute\\path\\to\\systemlog.cpp(22) : warning C4100: 'op' : unreferenced formal parameter",
            '2>------ Build started: Project: product, Configuration: debug Win32 ------',
        ]
        self.assertResult(
            nbFiles=1,
            nbErrors=1,
            nbProjects=2,
            nbWarnings=6,
            progress={'files': 1, 'projects': 2, 'warnings': 6},
            errors=errors,
            warnings=warnings,
        )


class VCx(vstudio.VisualStudio):
    def run(self):
        self.command = ["command", "here"]
        return super().run()


class VisualStudio(TestBuildStepMixin, TestReactorMixin, unittest.TestCase):
    """
    Test L{VisualStudio} with a simple subclass, L{VCx}.
    """

    def setUp(self):
        self.setup_test_reactor()
        return self.setup_test_build_step()

    def test_default_config(self):
        vs = vstudio.VisualStudio()
        self.assertEqual(vs.config, 'release')

    def test_simple(self):
        self.setup_step(VCx())
        self.expect_commands(ExpectShell(workdir='wkdir', command=['command', 'here']).exit(0))
        self.expect_outcome(result=SUCCESS, state_string="compile 0 projects 0 files")
        return self.run_step()

    def test_skipped(self):
        self.setup_step(VCx(doStepIf=False))
        self.expect_commands()
        self.expect_outcome(result=SKIPPED, state_string="")
        return self.run_step()

    @defer.inlineCallbacks
    def test_installdir(self):
        self.setup_step(VCx(installdir=r'C:\I'))
        self.get_nth_step(0).exp_installdir = r'C:\I'
        self.expect_commands(ExpectShell(workdir='wkdir', command=['command', 'here']).exit(0))
        self.expect_outcome(result=SUCCESS, state_string="compile 0 projects 0 files")
        yield self.run_step()
        self.assertEqual(self.get_nth_step(0).installdir, r'C:\I')

    def test_evaluate_result_failure(self):
        self.setup_step(VCx())
        self.expect_commands(ExpectShell(workdir='wkdir', command=['command', 'here']).exit(1))
        self.expect_outcome(result=FAILURE, state_string="compile 0 projects 0 files (failure)")
        return self.run_step()

    def test_evaluate_result_errors(self):
        self.setup_step(VCx())
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['command', 'here'])
            .stdout('error ABC123: foo\r\n')
            .exit(0)
        )
        self.expect_outcome(
            result=FAILURE, state_string="compile 0 projects 0 files 1 errors (failure)"
        )
        return self.run_step()

    def test_evaluate_result_warnings(self):
        self.setup_step(VCx())
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['command', 'here'])
            .stdout('foo: warning ABC123: foo\r\n')
            .exit(0)
        )
        self.expect_outcome(
            result=WARNINGS, state_string="compile 0 projects 0 files 1 warnings (warnings)"
        )
        return self.run_step()

    def test_env_setup(self):
        self.setup_step(
            VCx(
                INCLUDE=[r'c:\INC1', r'c:\INC2'],
                LIB=[r'c:\LIB1', r'C:\LIB2'],
                PATH=[r'c:\P1', r'C:\P2'],
            )
        )
        self.expect_commands(
            ExpectShell(
                workdir="wkdir",
                command=["command", "here"],
                env={
                    "INCLUDE": r"c:\INC1;c:\INC2;",
                    "LIB": r"c:\LIB1;C:\LIB2;",
                    "PATH": r"c:\P1;C:\P2;",
                },
            ).exit(0)
        )
        self.expect_outcome(result=SUCCESS, state_string="compile 0 projects 0 files")
        return self.run_step()

    def test_env_setup_existing(self):
        self.setup_step(
            VCx(
                INCLUDE=[r'c:\INC1', r'c:\INC2'],
                LIB=[r'c:\LIB1', r'C:\LIB2'],
                PATH=[r'c:\P1', r'C:\P2'],
            )
        )
        self.expect_commands(
            ExpectShell(
                workdir="wkdir",
                command=["command", "here"],
                env={
                    "INCLUDE": r"c:\INC1;c:\INC2;",
                    "LIB": r"c:\LIB1;C:\LIB2;",
                    "PATH": r"c:\P1;C:\P2;",
                },
            ).exit(0)
        )
        self.expect_outcome(result=SUCCESS, state_string="compile 0 projects 0 files")
        return self.run_step()

    @defer.inlineCallbacks
    def test_rendering(self):
        self.setup_step(VCx(projectfile=Property('a'), config=Property('b'), project=Property('c')))
        self.build.setProperty('a', 'aa', 'Test')
        self.build.setProperty('b', 'bb', 'Test')
        self.build.setProperty('c', 'cc', 'Test')
        self.expect_commands(ExpectShell(workdir='wkdir', command=['command', 'here']).exit(0))
        self.expect_outcome(result=SUCCESS, state_string="compile 0 projects 0 files")
        yield self.run_step()

        step = self.get_nth_step(0)
        self.assertEqual([step.projectfile, step.config, step.project], ['aa', 'bb', 'cc'])


class TestVC6(TestBuildStepMixin, TestReactorMixin, unittest.TestCase):
    def setUp(self):
        self.setup_test_reactor()
        return self.setup_test_build_step()

    def getExpectedEnv(self, installdir, LIB=None, p=None, i=None):
        include = [
            installdir + r'\VC98\INCLUDE;',
            installdir + r'\VC98\ATL\INCLUDE;',
            installdir + r'\VC98\MFC\INCLUDE;',
        ]
        lib = [
            installdir + r'\VC98\LIB;',
            installdir + r'\VC98\MFC\LIB;',
        ]
        path = [
            installdir + r'\Common\msdev98\BIN;',
            installdir + r'\VC98\BIN;',
            installdir + r'\Common\TOOLS\WINNT;',
            installdir + r'\Common\TOOLS;',
        ]
        if p:
            path.insert(0, f'{p};')
        if i:
            include.insert(0, f'{i};')
        if LIB:
            lib.insert(0, f'{LIB};')
        return {
            "INCLUDE": ''.join(include),
            "LIB": ''.join(lib),
            "PATH": ''.join(path),
        }

    def test_args(self):
        self.setup_step(vstudio.VC6(projectfile='pf', config='cfg', project='pj'))
        self.expect_commands(
            ExpectShell(
                workdir='wkdir',
                command=['msdev', 'pf', '/MAKE', 'pj - cfg', '/REBUILD'],
                env=self.getExpectedEnv(r'C:\Program Files\Microsoft Visual Studio'),
            ).exit(0)
        )
        self.expect_outcome(result=SUCCESS, state_string="compile 0 projects 0 files")
        return self.run_step()

    def test_clean(self):
        self.setup_step(vstudio.VC6(projectfile='pf', config='cfg', project='pj', mode='clean'))
        self.expect_commands(
            ExpectShell(
                workdir='wkdir',
                command=['msdev', 'pf', '/MAKE', 'pj - cfg', '/CLEAN'],
                env=self.getExpectedEnv(r'C:\Program Files\Microsoft Visual Studio'),
            ).exit(0)
        )
        self.expect_outcome(result=SUCCESS, state_string="compile 0 projects 0 files")
        return self.run_step()

    def test_noproj_build(self):
        self.setup_step(vstudio.VC6(projectfile='pf', config='cfg', mode='build'))
        self.expect_commands(
            ExpectShell(
                workdir='wkdir',
                command=['msdev', 'pf', '/MAKE', 'ALL - cfg', '/BUILD'],
                env=self.getExpectedEnv(r'C:\Program Files\Microsoft Visual Studio'),
            ).exit(0)
        )
        self.expect_outcome(result=SUCCESS, state_string="compile 0 projects 0 files")
        return self.run_step()

    def test_env_prepend(self):
        self.setup_step(
            vstudio.VC6(
                projectfile='pf', config='cfg', project='pj', PATH=['p'], INCLUDE=['i'], LIB=['l']
            )
        )
        self.expect_commands(
            ExpectShell(
                workdir='wkdir',
                command=[
                    'msdev',
                    'pf',
                    '/MAKE',
                    'pj - cfg',
                    '/REBUILD',
                    '/USEENV',
                ],  # note extra param
                env=self.getExpectedEnv(
                    r'C:\Program Files\Microsoft Visual Studio', LIB='l', p='p', i='i'
                ),
            ).exit(0)
        )
        self.expect_outcome(result=SUCCESS, state_string="compile 0 projects 0 files")
        return self.run_step()


class TestVC7(TestBuildStepMixin, TestReactorMixin, unittest.TestCase):
    def setUp(self):
        self.setup_test_reactor()
        return self.setup_test_build_step()

    def getExpectedEnv(self, installdir, LIB=None, p=None, i=None):
        include = [
            installdir + r'\VC7\INCLUDE;',
            installdir + r'\VC7\ATLMFC\INCLUDE;',
            installdir + r'\VC7\PlatformSDK\include;',
            installdir + r'\SDK\v1.1\include;',
        ]
        lib = [
            installdir + r'\VC7\LIB;',
            installdir + r'\VC7\ATLMFC\LIB;',
            installdir + r'\VC7\PlatformSDK\lib;',
            installdir + r'\SDK\v1.1\lib;',
        ]
        path = [
            installdir + r'\Common7\IDE;',
            installdir + r'\VC7\BIN;',
            installdir + r'\Common7\Tools;',
            installdir + r'\Common7\Tools\bin;',
        ]
        if p:
            path.insert(0, f'{p};')
        if i:
            include.insert(0, f'{i};')
        if LIB:
            lib.insert(0, f'{LIB};')
        return {
            "INCLUDE": ''.join(include),
            "LIB": ''.join(lib),
            "PATH": ''.join(path),
        }

    def test_args(self):
        self.setup_step(vstudio.VC7(projectfile='pf', config='cfg', project='pj'))
        self.expect_commands(
            ExpectShell(
                workdir='wkdir',
                command=['devenv.com', 'pf', '/Rebuild', 'cfg', '/Project', 'pj'],
                env=self.getExpectedEnv(r'C:\Program Files\Microsoft Visual Studio .NET 2003'),
            ).exit(0)
        )
        self.expect_outcome(result=SUCCESS, state_string="compile 0 projects 0 files")
        return self.run_step()

    def test_clean(self):
        self.setup_step(vstudio.VC7(projectfile='pf', config='cfg', project='pj', mode='clean'))
        self.expect_commands(
            ExpectShell(
                workdir='wkdir',
                command=['devenv.com', 'pf', '/Clean', 'cfg', '/Project', 'pj'],
                env=self.getExpectedEnv(r'C:\Program Files\Microsoft Visual Studio .NET 2003'),
            ).exit(0)
        )
        self.expect_outcome(result=SUCCESS, state_string="compile 0 projects 0 files")
        return self.run_step()

    def test_noproj_build(self):
        self.setup_step(vstudio.VC7(projectfile='pf', config='cfg', mode='build'))
        self.expect_commands(
            ExpectShell(
                workdir='wkdir',
                command=['devenv.com', 'pf', '/Build', 'cfg'],
                env=self.getExpectedEnv(r'C:\Program Files\Microsoft Visual Studio .NET 2003'),
            ).exit(0)
        )
        self.expect_outcome(result=SUCCESS, state_string="compile 0 projects 0 files")
        return self.run_step()

    def test_env_prepend(self):
        self.setup_step(
            vstudio.VC7(
                projectfile='pf', config='cfg', project='pj', PATH=['p'], INCLUDE=['i'], LIB=['l']
            )
        )
        self.expect_commands(
            ExpectShell(
                workdir='wkdir',
                command=['devenv.com', 'pf', '/Rebuild', 'cfg', '/UseEnv', '/Project', 'pj'],
                env=self.getExpectedEnv(
                    r'C:\Program Files\Microsoft Visual Studio .NET 2003', LIB='l', p='p', i='i'
                ),
            ).exit(0)
        )
        self.expect_outcome(result=SUCCESS, state_string="compile 0 projects 0 files")
        return self.run_step()


class VC8ExpectedEnvMixin:
    # used for VC8 and VC9Express

    def getExpectedEnv(self, installdir, x64=False, LIB=None, i=None, p=None):
        include = [
            installdir + r'\VC\INCLUDE;',
            installdir + r'\VC\ATLMFC\include;',
            installdir + r'\VC\PlatformSDK\include;',
        ]
        lib = [
            installdir + r'\VC\LIB;',
            installdir + r'\VC\ATLMFC\LIB;',
            installdir + r'\VC\PlatformSDK\lib;',
            installdir + r'\SDK\v2.0\lib;',
        ]
        path = [
            installdir + r'\Common7\IDE;',
            installdir + r'\VC\BIN;',
            installdir + r'\Common7\Tools;',
            installdir + r'\Common7\Tools\bin;',
            installdir + r'\VC\PlatformSDK\bin;',
            installdir + r'\SDK\v2.0\bin;',
            installdir + r'\VC\VCPackages;',
            r'${PATH};',
        ]
        if x64:
            path.insert(1, installdir + r'\VC\BIN\x86_amd64;')
            lib = [lb[:-1] + r'\amd64;' for lb in lib]
        if LIB:
            lib.insert(0, f'{LIB};')
        if p:
            path.insert(0, f'{p};')
        if i:
            include.insert(0, f'{i};')
        return {
            "INCLUDE": ''.join(include),
            "LIB": ''.join(lib),
            "PATH": ''.join(path),
        }


class TestVC8(VC8ExpectedEnvMixin, TestBuildStepMixin, TestReactorMixin, unittest.TestCase):
    def setUp(self):
        self.setup_test_reactor()
        return self.setup_test_build_step()

    def test_args(self):
        self.setup_step(vstudio.VC8(projectfile='pf', config='cfg', project='pj', arch='arch'))
        self.expect_commands(
            ExpectShell(
                workdir='wkdir',
                command=['devenv.com', 'pf', '/Rebuild', 'cfg', '/Project', 'pj'],
                env=self.getExpectedEnv(r'C:\Program Files\Microsoft Visual Studio 8'),
            ).exit(0)
        )
        self.expect_outcome(result=SUCCESS, state_string="compile 0 projects 0 files")
        return self.run_step()

    def test_args_x64(self):
        self.setup_step(vstudio.VC8(projectfile='pf', config='cfg', project='pj', arch='x64'))
        self.expect_commands(
            ExpectShell(
                workdir='wkdir',
                command=['devenv.com', 'pf', '/Rebuild', 'cfg', '/Project', 'pj'],
                env=self.getExpectedEnv(r'C:\Program Files\Microsoft Visual Studio 8', x64=True),
            ).exit(0)
        )
        self.expect_outcome(result=SUCCESS, state_string="compile 0 projects 0 files")
        return self.run_step()

    def test_clean(self):
        self.setup_step(vstudio.VC8(projectfile='pf', config='cfg', project='pj', mode='clean'))
        self.expect_commands(
            ExpectShell(
                workdir='wkdir',
                command=['devenv.com', 'pf', '/Clean', 'cfg', '/Project', 'pj'],
                env=self.getExpectedEnv(r'C:\Program Files\Microsoft Visual Studio 8'),
            ).exit(0)
        )
        self.expect_outcome(result=SUCCESS, state_string="compile 0 projects 0 files")
        return self.run_step()

    @defer.inlineCallbacks
    def test_rendering(self):
        self.setup_step(vstudio.VC8(projectfile='pf', config='cfg', arch=Property('a')))
        self.build.setProperty('a', 'x64', 'Test')
        self.expect_commands(
            ExpectShell(
                workdir='wkdir',
                command=['devenv.com', 'pf', '/Rebuild', 'cfg'],
                env=self.getExpectedEnv(r'C:\Program Files\Microsoft Visual Studio 8', x64=True),
            ).exit(0)  # property has expected effect
        )
        self.expect_outcome(result=SUCCESS, state_string="compile 0 projects 0 files")
        yield self.run_step()

        self.assertEqual(self.get_nth_step(0).arch, 'x64')


class TestVCExpress9(VC8ExpectedEnvMixin, TestBuildStepMixin, TestReactorMixin, unittest.TestCase):
    def setUp(self):
        self.setup_test_reactor()
        return self.setup_test_build_step()

    def test_args(self):
        self.setup_step(vstudio.VCExpress9(projectfile='pf', config='cfg', project='pj'))
        self.expect_commands(
            ExpectShell(
                workdir='wkdir',
                command=['vcexpress', 'pf', '/Rebuild', 'cfg', '/Project', 'pj'],
                env=self.getExpectedEnv(
                    # note: still uses version 8 (?!)
                    r'C:\Program Files\Microsoft Visual Studio 8'
                ),
            ).exit(0)
        )
        self.expect_outcome(result=SUCCESS, state_string="compile 0 projects 0 files")
        return self.run_step()

    def test_clean(self):
        self.setup_step(
            vstudio.VCExpress9(projectfile='pf', config='cfg', project='pj', mode='clean')
        )
        self.expect_commands(
            ExpectShell(
                workdir='wkdir',
                command=['vcexpress', 'pf', '/Clean', 'cfg', '/Project', 'pj'],
                env=self.getExpectedEnv(
                    # note: still uses version 8 (?!)
                    r'C:\Program Files\Microsoft Visual Studio 8'
                ),
            ).exit(0)
        )
        self.expect_outcome(result=SUCCESS, state_string="compile 0 projects 0 files")
        return self.run_step()

    def test_mode_build_env(self):
        self.setup_step(
            vstudio.VCExpress9(
                projectfile='pf', config='cfg', project='pj', mode='build', INCLUDE=['i']
            )
        )
        self.expect_commands(
            ExpectShell(
                workdir='wkdir',
                command=['vcexpress', 'pf', '/Build', 'cfg', '/UseEnv', '/Project', 'pj'],
                env=self.getExpectedEnv(r'C:\Program Files\Microsoft Visual Studio 8', i='i'),
            ).exit(0)
        )
        self.expect_outcome(result=SUCCESS, state_string="compile 0 projects 0 files")
        return self.run_step()


class TestVC9(VC8ExpectedEnvMixin, TestBuildStepMixin, TestReactorMixin, unittest.TestCase):
    def setUp(self):
        self.setup_test_reactor()
        return self.setup_test_build_step()

    def test_installdir(self):
        self.setup_step(vstudio.VC9(projectfile='pf', config='cfg', project='pj'))
        self.expect_commands(
            ExpectShell(
                workdir='wkdir',
                command=['devenv.com', 'pf', '/Rebuild', 'cfg', '/Project', 'pj'],
                env=self.getExpectedEnv(r'C:\Program Files\Microsoft Visual Studio 9.0'),
            ).exit(0)
        )
        self.expect_outcome(result=SUCCESS, state_string="compile 0 projects 0 files")
        return self.run_step()


class TestVC10(VC8ExpectedEnvMixin, TestBuildStepMixin, TestReactorMixin, unittest.TestCase):
    def setUp(self):
        self.setup_test_reactor()
        return self.setup_test_build_step()

    def test_installdir(self):
        self.setup_step(vstudio.VC10(projectfile='pf', config='cfg', project='pj'))
        self.expect_commands(
            ExpectShell(
                workdir='wkdir',
                command=['devenv.com', 'pf', '/Rebuild', 'cfg', '/Project', 'pj'],
                env=self.getExpectedEnv(r'C:\Program Files\Microsoft Visual Studio 10.0'),
            ).exit(0)
        )
        self.expect_outcome(result=SUCCESS, state_string="compile 0 projects 0 files")
        return self.run_step()


class TestVC11(VC8ExpectedEnvMixin, TestBuildStepMixin, TestReactorMixin, unittest.TestCase):
    def setUp(self):
        self.setup_test_reactor()
        return self.setup_test_build_step()

    def test_installdir(self):
        self.setup_step(vstudio.VC11(projectfile='pf', config='cfg', project='pj'))
        self.expect_commands(
            ExpectShell(
                workdir='wkdir',
                command=['devenv.com', 'pf', '/Rebuild', 'cfg', '/Project', 'pj'],
                env=self.getExpectedEnv(r'C:\Program Files\Microsoft Visual Studio 11.0'),
            ).exit(0)
        )
        self.expect_outcome(result=SUCCESS, state_string="compile 0 projects 0 files")
        return self.run_step()


class TestMsBuild(TestBuildStepMixin, TestReactorMixin, unittest.TestCase):
    def setUp(self):
        self.setup_test_reactor()
        return self.setup_test_build_step()

    @defer.inlineCallbacks
    def test_no_platform(self):
        self.setup_step(
            vstudio.MsBuild(projectfile='pf', config='cfg', platform=None, project='pj')
        )

        self.expect_outcome(result=results.EXCEPTION, state_string="built pj for cfg|None")
        yield self.run_step()
        self.assertEqual(len(self.flushLoggedErrors(config.ConfigErrors)), 1)

    def test_rebuild_project(self):
        self.setup_step(
            vstudio.MsBuild(projectfile='pf', config='cfg', platform='Win32', project='pj')
        )

        self.expect_commands(
            ExpectShell(
                workdir='wkdir',
                command='"%VCENV_BAT%" x86 && msbuild "pf" /p:Configuration="cfg" /p:Platform="Win32" /maxcpucount /t:"pj"',
                env={'VCENV_BAT': r'${VS110COMNTOOLS}..\..\VC\vcvarsall.bat'},
            ).exit(0)
        )
        self.expect_outcome(result=SUCCESS, state_string="built pj for cfg|Win32")
        return self.run_step()

    def test_build_project(self):
        self.setup_step(
            vstudio.MsBuild(
                projectfile='pf', config='cfg', platform='Win32', project='pj', mode='build'
            )
        )

        self.expect_commands(
            ExpectShell(
                workdir='wkdir',
                command='"%VCENV_BAT%" x86 && msbuild "pf" /p:Configuration="cfg" /p:Platform="Win32" /maxcpucount /t:"pj:Build"',
                env={'VCENV_BAT': r'${VS110COMNTOOLS}..\..\VC\vcvarsall.bat'},
            ).exit(0)
        )
        self.expect_outcome(result=SUCCESS, state_string="built pj for cfg|Win32")
        return self.run_step()

    def test_clean_project(self):
        self.setup_step(
            vstudio.MsBuild(
                projectfile='pf', config='cfg', platform='Win32', project='pj', mode='clean'
            )
        )

        self.expect_commands(
            ExpectShell(
                workdir='wkdir',
                command='"%VCENV_BAT%" x86 && msbuild "pf" /p:Configuration="cfg" /p:Platform="Win32" /maxcpucount /t:"pj:Clean"',
                env={'VCENV_BAT': r'${VS110COMNTOOLS}..\..\VC\vcvarsall.bat'},
            ).exit(0)
        )
        self.expect_outcome(result=SUCCESS, state_string="built pj for cfg|Win32")
        return self.run_step()

    def test_rebuild_project_with_defines(self):
        self.setup_step(
            vstudio.MsBuild(
                projectfile='pf',
                config='cfg',
                platform='Win32',
                project='pj',
                defines=['Define1', 'Define2=42'],
            )
        )

        self.expect_commands(
            ExpectShell(
                workdir='wkdir',
                command='"%VCENV_BAT%" x86 && msbuild "pf" /p:Configuration="cfg" /p:Platform="Win32" /maxcpucount /t:"pj" /p:DefineConstants="Define1;Define2=42"',
                env={'VCENV_BAT': r'${VS110COMNTOOLS}..\..\VC\vcvarsall.bat'},
            ).exit(0)
        )
        self.expect_outcome(result=SUCCESS, state_string="built pj for cfg|Win32")
        return self.run_step()

    def test_rebuild_solution(self):
        self.setup_step(vstudio.MsBuild(projectfile='pf', config='cfg', platform='x64'))

        self.expect_commands(
            ExpectShell(
                workdir='wkdir',
                command='"%VCENV_BAT%" x86 && msbuild "pf" /p:Configuration="cfg" /p:Platform="x64" /maxcpucount /t:Rebuild',
                env={'VCENV_BAT': r'${VS110COMNTOOLS}..\..\VC\vcvarsall.bat'},
            ).exit(0)
        )
        self.expect_outcome(result=SUCCESS, state_string="built solution for cfg|x64")
        return self.run_step()


class TestMsBuild141(TestBuildStepMixin, TestReactorMixin, unittest.TestCase):
    def setUp(self):
        self.setup_test_reactor()
        return self.setup_test_build_step()

    @defer.inlineCallbacks
    def test_no_platform(self):
        self.setup_step(
            vstudio.MsBuild(projectfile='pf', config='cfg', platform=None, project='pj')
        )

        self.expect_outcome(result=results.EXCEPTION, state_string="built pj for cfg|None")
        yield self.run_step()
        self.assertEqual(len(self.flushLoggedErrors(config.ConfigErrors)), 1)

    def test_rebuild_project(self):
        self.setup_step(
            vstudio.MsBuild141(projectfile='pf', config='cfg', platform='Win32', project='pj')
        )

        self.expect_commands(
            ExpectShell(
                workdir='wkdir',
                command='FOR /F "tokens=*" %%I in (\'vswhere.exe -version "[15.0,16.0)" -products * -property installationPath\')  do "%%I\\%VCENV_BAT%" x86 && msbuild "pf" /p:Configuration="cfg" /p:Platform="Win32" /maxcpucount /t:"pj"',
                env={
                    'VCENV_BAT': r'\VC\Auxiliary\Build\vcvarsall.bat',
                    'PATH': 'C:\\Program Files (x86)\\Microsoft Visual Studio\\Installer\\;C:\\Program Files\\Microsoft Visual Studio\\Installer\\;${PATH};',
                },
            ).exit(0)
        )
        self.expect_outcome(result=SUCCESS, state_string="compile 0 projects 0 files")
        return self.run_step()

    def test_build_project(self):
        self.setup_step(
            vstudio.MsBuild141(
                projectfile='pf', config='cfg', platform='Win32', project='pj', mode='build'
            )
        )

        self.expect_commands(
            ExpectShell(
                workdir='wkdir',
                command='FOR /F "tokens=*" %%I in (\'vswhere.exe -version "[15.0,16.0)" -products * -property installationPath\')  do "%%I\\%VCENV_BAT%" x86 && msbuild "pf" /p:Configuration="cfg" /p:Platform="Win32" /maxcpucount /t:"pj:Build"',
                env={
                    'VCENV_BAT': r'\VC\Auxiliary\Build\vcvarsall.bat',
                    'PATH': 'C:\\Program Files (x86)\\Microsoft Visual Studio\\Installer\\;C:\\Program Files\\Microsoft Visual Studio\\Installer\\;${PATH};',
                },
            ).exit(0)
        )
        self.expect_outcome(result=SUCCESS, state_string="compile 0 projects 0 files")
        return self.run_step()

    def test_clean_project(self):
        self.setup_step(
            vstudio.MsBuild141(
                projectfile='pf', config='cfg', platform='Win32', project='pj', mode='clean'
            )
        )

        self.expect_commands(
            ExpectShell(
                workdir='wkdir',
                command='FOR /F "tokens=*" %%I in (\'vswhere.exe -version "[15.0,16.0)" -products * -property installationPath\')  do "%%I\\%VCENV_BAT%" x86 && msbuild "pf" /p:Configuration="cfg" /p:Platform="Win32" /maxcpucount /t:"pj:Clean"',
                env={
                    'VCENV_BAT': r'\VC\Auxiliary\Build\vcvarsall.bat',
                    'PATH': 'C:\\Program Files (x86)\\Microsoft Visual Studio\\Installer\\;C:\\Program Files\\Microsoft Visual Studio\\Installer\\;${PATH};',
                },
            ).exit(0)
        )
        self.expect_outcome(result=SUCCESS, state_string="compile 0 projects 0 files")
        return self.run_step()

    def test_rebuild_project_with_defines(self):
        self.setup_step(
            vstudio.MsBuild141(
                projectfile='pf',
                config='cfg',
                platform='Win32',
                project='pj',
                defines=['Define1', 'Define2=42'],
            )
        )

        self.expect_commands(
            ExpectShell(
                workdir='wkdir',
                command='FOR /F "tokens=*" %%I in (\'vswhere.exe -version "[15.0,16.0)" -products * -property installationPath\')  do "%%I\\%VCENV_BAT%" x86 && msbuild "pf" /p:Configuration="cfg" /p:Platform="Win32" /maxcpucount /t:"pj" /p:DefineConstants="Define1;Define2=42"',
                env={
                    'VCENV_BAT': r'\VC\Auxiliary\Build\vcvarsall.bat',
                    'PATH': 'C:\\Program Files (x86)\\Microsoft Visual Studio\\Installer\\;C:\\Program Files\\Microsoft Visual Studio\\Installer\\;${PATH};',
                },
            ).exit(0)
        )
        self.expect_outcome(result=SUCCESS, state_string="compile 0 projects 0 files")
        return self.run_step()

    def test_rebuild_solution(self):
        self.setup_step(vstudio.MsBuild141(projectfile='pf', config='cfg', platform='x64'))

        self.expect_commands(
            ExpectShell(
                workdir='wkdir',
                command='FOR /F "tokens=*" %%I in (\'vswhere.exe -version "[15.0,16.0)" -products * -property installationPath\')  do "%%I\\%VCENV_BAT%" x86 && msbuild "pf" /p:Configuration="cfg" /p:Platform="x64" /maxcpucount /t:Rebuild',
                env={
                    'VCENV_BAT': r'\VC\Auxiliary\Build\vcvarsall.bat',
                    'PATH': 'C:\\Program Files (x86)\\Microsoft Visual Studio\\Installer\\;C:\\Program Files\\Microsoft Visual Studio\\Installer\\;${PATH};',
                },
            ).exit(0)
        )
        self.expect_outcome(result=SUCCESS, state_string="compile 0 projects 0 files")
        return self.run_step()

    def test_aliases_MsBuild15(self):
        self.assertIdentical(vstudio.MsBuild141, vstudio.MsBuild15)


class TestMsBuild16(TestBuildStepMixin, TestReactorMixin, unittest.TestCase):
    def setUp(self):
        self.setup_test_reactor()
        return self.setup_test_build_step()

    def test_version_range_is_correct(self):
        self.setup_step(
            vstudio.MsBuild16(projectfile='pf', config='cfg', platform='Win32', project='pj')
        )

        self.expect_commands(
            ExpectShell(
                workdir='wkdir',
                command='FOR /F "tokens=*" %%I in (\'vswhere.exe -version "[16.0,17.0)" -products * -property installationPath\')  do "%%I\\%VCENV_BAT%" x86 && msbuild "pf" /p:Configuration="cfg" /p:Platform="Win32" /maxcpucount /t:"pj"',
                env={
                    'VCENV_BAT': r'\VC\Auxiliary\Build\vcvarsall.bat',
                    'PATH': 'C:\\Program Files (x86)\\Microsoft Visual Studio\\Installer\\;C:\\Program Files\\Microsoft Visual Studio\\Installer\\;${PATH};',
                },
            ).exit(0)
        )
        self.expect_outcome(result=SUCCESS, state_string="compile 0 projects 0 files")
        return self.run_step()


class TestMsBuild17(TestBuildStepMixin, TestReactorMixin, unittest.TestCase):
    def setUp(self):
        self.setup_test_reactor()
        return self.setup_test_build_step()

    def test_version_range_is_correct(self):
        self.setup_step(
            vstudio.MsBuild17(projectfile='pf', config='cfg', platform='Win32', project='pj')
        )

        self.expect_commands(
            ExpectShell(
                workdir='wkdir',
                command='FOR /F "tokens=*" %%I in (\'vswhere.exe -version "[17.0,18.0)" -products * -property installationPath\')  do "%%I\\%VCENV_BAT%" x86 && msbuild "pf" /p:Configuration="cfg" /p:Platform="Win32" /maxcpucount /t:"pj"',
                env={
                    'VCENV_BAT': r'\VC\Auxiliary\Build\vcvarsall.bat',
                    'PATH': 'C:\\Program Files (x86)\\Microsoft Visual Studio\\Installer\\;C:\\Program Files\\Microsoft Visual Studio\\Installer\\;${PATH};',
                },
            ).exit(0)
        )
        self.expect_outcome(result=SUCCESS, state_string="compile 0 projects 0 files")
        return self.run_step()


class Aliases(unittest.TestCase):
    def test_vs2003(self):
        self.assertIdentical(vstudio.VS2003, vstudio.VC7)

    def test_vs2005(self):
        self.assertIdentical(vstudio.VS2005, vstudio.VC8)

    def test_vs2008(self):
        self.assertIdentical(vstudio.VS2008, vstudio.VC9)

    def test_vs2010(self):
        self.assertIdentical(vstudio.VS2010, vstudio.VC10)

    def test_vs2012(self):
        self.assertIdentical(vstudio.VS2012, vstudio.VC11)



from twisted.internet import defer
from twisted.trial import unittest

from buildbot.secrets.manager import SecretManager
from buildbot.secrets.secret import SecretDetails
from buildbot.test.fake import fakemaster
from buildbot.test.fake.secrets import FakeSecretStorage
from buildbot.test.reactor import TestReactorMixin


class TestSecretsManager(TestReactorMixin, unittest.TestCase):
    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.master = yield fakemaster.make_master(self)
        self.master.config.secretsProviders = [
            FakeSecretStorage(secretdict={"foo": "bar", "other": "value"})
        ]

    @defer.inlineCallbacks
    def testGetManagerService(self):
        secret_service_manager = SecretManager()
        fakeStorageService = FakeSecretStorage()
        fakeStorageService.reconfigService(secretdict={"foo": "bar", "other": "value"})
        secret_service_manager.services = [fakeStorageService]
        expectedClassName = FakeSecretStorage.__name__
        expectedSecretDetail = SecretDetails(expectedClassName, "foo", "bar")
        secret_result = yield secret_service_manager.get("foo")
        strExpectedSecretDetail = str(secret_result)
        self.assertEqual(secret_result, expectedSecretDetail)
        self.assertEqual(secret_result.key, "foo")
        self.assertEqual(secret_result.value, "bar")
        self.assertEqual(secret_result.source, expectedClassName)
        self.assertEqual(strExpectedSecretDetail, "FakeSecretStorage foo: 'bar'")

    @defer.inlineCallbacks
    def testGetNoDataManagerService(self):
        secret_service_manager = SecretManager()
        fakeStorageService = FakeSecretStorage()
        fakeStorageService.reconfigService(secretdict={"foo": "bar", "other": "value"})
        secret_service_manager.services = [fakeStorageService]
        secret_result = yield secret_service_manager.get("foo2")
        self.assertEqual(secret_result, None)

    @defer.inlineCallbacks
    def testGetDataMultipleManagerService(self):
        secret_service_manager = SecretManager()
        fakeStorageService = FakeSecretStorage()
        fakeStorageService.reconfigService(secretdict={"foo": "bar", "other": "value"})
        otherFakeStorageService = FakeSecretStorage()
        otherFakeStorageService.reconfigService(secretdict={"foo2": "bar", "other2": "value"})

        secret_service_manager.services = [fakeStorageService, otherFakeStorageService]
        expectedSecretDetail = SecretDetails(FakeSecretStorage.__name__, "foo2", "bar")
        secret_result = yield secret_service_manager.get("foo2")
        self.assertEqual(secret_result, expectedSecretDetail)

    @defer.inlineCallbacks
    def testGetDataMultipleManagerValues(self):
        secret_service_manager = SecretManager()
        fakeStorageService = FakeSecretStorage()
        fakeStorageService.reconfigService(secretdict={"foo": "bar", "other": ""})
        otherFakeStorageService = FakeSecretStorage()
        otherFakeStorageService.reconfigService(secretdict={"foo2": "bar2", "other": ""})

        secret_service_manager.services = [fakeStorageService, otherFakeStorageService]
        expectedSecretDetail = SecretDetails(FakeSecretStorage.__name__, "other", "")
        secret_result = yield secret_service_manager.get("other")
        self.assertEqual(secret_result, expectedSecretDetail)

    @defer.inlineCallbacks
    def testGetDataMultipleManagerServiceNoDatas(self):
        secret_service_manager = SecretManager()
        fakeStorageService = FakeSecretStorage()
        fakeStorageService.reconfigService(secretdict={"foo": "bar", "other": "value"})
        otherFakeStorageService = FakeSecretStorage()
        otherFakeStorageService.reconfigService(secretdict={"foo2": "bar", "other2": "value"})
        secret_service_manager.services = [fakeStorageService, otherFakeStorageService]
        secret_result = yield secret_service_manager.get("foo3")
        self.assertEqual(secret_result, None)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import os
import shutil
import stat
import sys
import tempfile


class PrivateTemporaryDirectory:
    """Works similarly to python 3.2+ TemporaryDirectory except the
    also sets the permissions of the created directory and

    Note, that Windows ignores the permissions.
    """

    def __init__(self, suffix=None, prefix=None, dir=None, mode=0o700):
        self.name = tempfile.mkdtemp(suffix, prefix, dir)
        self.mode = mode
        self._cleanup_needed = True

    def __enter__(self):
        return self.name

    def __exit__(self, exc, value, tb):
        self.cleanup()

    def cleanup(self):
        if self._cleanup_needed:

            def remove_readonly(func, path, _):
                """Workaround Permission Error on Windows if any files in path are read-only.

                See https://docs.python.org/3/library/shutil.html#rmtree-example
                """
                os.chmod(path, stat.S_IWRITE)
                func(path)

            if sys.version_info >= (3, 12):
                shutil.rmtree(self.name, onexc=remove_readonly)
            else:
                shutil.rmtree(self.name, onerror=remove_readonly)
            self._cleanup_needed = False



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import sqlalchemy as sa
from sqlalchemy.engine import url
from sqlalchemy.pool import NullPool
from twisted.python import runtime
from twisted.trial import unittest

from buildbot.db import enginestrategy
from buildbot.util.sautils import sa_version


class BuildbotCreateEngineTest(unittest.TestCase):
    "Test the special case methods, without actually creating a db"

    # used several times below
    mysql_kwargs = {
        "basedir": 'my-base-dir',
        "connect_args": {"init_command": 'SET default_storage_engine=MyISAM'},
        "pool_recycle": 3600,
    }
    sqlite_kwargs = {"basedir": '/my-base-dir', "poolclass": NullPool}

    # utility

    def filter_kwargs(self, kwargs):
        # filter out the listeners list to just include the class name
        if 'listeners' in kwargs:
            kwargs['listeners'] = [lstnr.__class__.__name__ for lstnr in kwargs['listeners']]
        return kwargs

    # tests

    def test_sqlite_pct_sub(self):
        u = url.make_url("sqlite:///%(basedir)s/x/state.sqlite")
        kwargs = {"basedir": '/my-base-dir'}
        u, kwargs, max_conns = enginestrategy.special_case_sqlite(u, kwargs)
        self.assertEqual(
            [str(u), max_conns, self.filter_kwargs(kwargs)],
            ["sqlite:////my-base-dir/x/state.sqlite", 1, self.sqlite_kwargs],
        )

    def test_sqlite_relpath(self):
        url_src = "sqlite:///x/state.sqlite"
        basedir = "/my-base-dir"
        expected_url = "sqlite:////my-base-dir/x/state.sqlite"

        # this looks a whole lot different on windows
        if runtime.platformType == 'win32':
            url_src = r'sqlite:///X\STATE.SQLITE'
            basedir = r'C:\MYBASE~1'
            expected_url = r'sqlite:///C:\MYBASE~1\X\STATE.SQLITE'

        exp_kwargs = self.sqlite_kwargs.copy()
        exp_kwargs['basedir'] = basedir

        u = url.make_url(url_src)
        kwargs = {"basedir": basedir}
        u, kwargs, max_conns = enginestrategy.special_case_sqlite(u, kwargs)
        self.assertEqual(
            [str(u), max_conns, self.filter_kwargs(kwargs)], [expected_url, 1, exp_kwargs]
        )

    def test_sqlite_abspath(self):
        u = url.make_url("sqlite:////x/state.sqlite")
        kwargs = {"basedir": '/my-base-dir'}
        u, kwargs, max_conns = enginestrategy.special_case_sqlite(u, kwargs)
        self.assertEqual(
            [str(u), max_conns, self.filter_kwargs(kwargs)],
            ["sqlite:////x/state.sqlite", 1, self.sqlite_kwargs],
        )

    def test_sqlite_memory(self):
        u = url.make_url("sqlite://")
        kwargs = {"basedir": 'my-base-dir'}
        u, kwargs, max_conns = enginestrategy.special_case_sqlite(u, kwargs)
        self.assertEqual(
            [str(u), max_conns, self.filter_kwargs(kwargs)],
            [
                "sqlite://",
                1,  # only one conn at a time
                {"basedir": 'my-base-dir', "connect_args": {"check_same_thread": False}},
            ],
        )

    def test_mysql_simple(self):
        u = url.make_url("mysql://host/dbname")
        kwargs = {"basedir": 'my-base-dir'}
        u, kwargs, max_conns = enginestrategy.special_case_mysql(u, kwargs)
        self.assertEqual(
            [str(u), max_conns, self.filter_kwargs(kwargs)],
            ["mysql://host/dbname?charset=utf8&use_unicode=True", None, self.mysql_kwargs],
        )

    def test_mysql_userport(self):
        u = url.make_url("mysql://user:pass@host:1234/dbname")
        kwargs = {"basedir": 'my-base-dir'}
        u, kwargs, max_conns = enginestrategy.special_case_mysql(u, kwargs)
        self.assertEqual(
            [str(u), max_conns, self.filter_kwargs(kwargs)],
            [
                (
                    "mysql://user:pass@host:1234/dbname?charset=utf8&use_unicode=True"
                    if sa_version()[0] < 2
                    else "mysql://user:***@host:1234/dbname?charset=utf8&use_unicode=True"
                ),
                None,
                self.mysql_kwargs,
            ],
        )

    def test_mysql_local(self):
        u = url.make_url("mysql:///dbname")
        kwargs = {"basedir": 'my-base-dir'}
        u, kwargs, max_conns = enginestrategy.special_case_mysql(u, kwargs)
        self.assertEqual(
            [str(u), max_conns, self.filter_kwargs(kwargs)],
            ["mysql:///dbname?charset=utf8&use_unicode=True", None, self.mysql_kwargs],
        )

    def test_mysql_args(self):
        u = url.make_url("mysql:///dbname?foo=bar")
        kwargs = {"basedir": 'my-base-dir'}
        u, kwargs, max_conns = enginestrategy.special_case_mysql(u, kwargs)
        self.assertEqual(
            [str(u), max_conns, self.filter_kwargs(kwargs)],
            ["mysql:///dbname?charset=utf8&foo=bar&use_unicode=True", None, self.mysql_kwargs],
        )

    def test_mysql_max_idle(self):
        u = url.make_url("mysql:///dbname?max_idle=1234")
        kwargs = {"basedir": 'my-base-dir'}
        u, kwargs, max_conns = enginestrategy.special_case_mysql(u, kwargs)
        exp = self.mysql_kwargs.copy()
        exp['pool_recycle'] = 1234
        self.assertEqual(
            [str(u), max_conns, self.filter_kwargs(kwargs)],
            ["mysql:///dbname?charset=utf8&use_unicode=True", None, exp],
        )

    def test_mysql_good_charset(self):
        u = url.make_url("mysql:///dbname?charset=utf8")
        kwargs = {"basedir": 'my-base-dir'}
        u, kwargs, max_conns = enginestrategy.special_case_mysql(u, kwargs)
        self.assertEqual(
            [str(u), max_conns, self.filter_kwargs(kwargs)],
            ["mysql:///dbname?charset=utf8&use_unicode=True", None, self.mysql_kwargs],
        )

    def test_mysql_bad_charset(self):
        u = url.make_url("mysql:///dbname?charset=ebcdic")
        kwargs = {"basedir": 'my-base-dir'}
        with self.assertRaises(TypeError):
            enginestrategy.special_case_mysql(u, kwargs)

    def test_mysql_good_use_unicode(self):
        u = url.make_url("mysql:///dbname?use_unicode=True")
        kwargs = {"basedir": 'my-base-dir'}
        u, kwargs, max_conns = enginestrategy.special_case_mysql(u, kwargs)
        self.assertEqual(
            [str(u), max_conns, self.filter_kwargs(kwargs)],
            ["mysql:///dbname?charset=utf8&use_unicode=True", None, self.mysql_kwargs],
        )

    def test_mysql_bad_use_unicode(self):
        u = url.make_url("mysql:///dbname?use_unicode=maybe")
        kwargs = {"basedir": 'my-base-dir'}
        with self.assertRaises(TypeError):
            enginestrategy.special_case_mysql(u, kwargs)

    def test_mysql_storage_engine(self):
        u = url.make_url("mysql:///dbname?storage_engine=foo")
        kwargs = {"basedir": 'my-base-dir'}
        u, kwargs, max_conns = enginestrategy.special_case_mysql(u, kwargs)
        exp = self.mysql_kwargs.copy()
        exp['connect_args'] = {"init_command": 'SET default_storage_engine=foo'}
        self.assertEqual(
            [str(u), max_conns, self.filter_kwargs(kwargs)],
            ["mysql:///dbname?charset=utf8&use_unicode=True", None, exp],
        )


class BuildbotEngineStrategy(unittest.TestCase):
    "Test create_engine by creating a sqlite in-memory db"

    def test_create_engine(self):
        engine = enginestrategy.create_engine('sqlite://', basedir="/base")
        try:
            with engine.connect() as conn:
                self.assertEqual(conn.scalar(sa.text("SELECT 13 + 14")), 27)
        finally:
            engine.dispose()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from parameterized import parameterized
from twisted.trial import unittest

from buildbot.test.util import config
from buildbot.util.git import GitMixin
from buildbot.util.git import check_ssh_config
from buildbot.util.git import ensureSshKeyNewline
from buildbot.util.git import escapeShellArgIfNeeded
from buildbot.util.git import getSshKnownHostsContents
from buildbot.util.git import scp_style_to_url_syntax


class TestEscapeShellArgIfNeeded(unittest.TestCase):
    def assert_escapes(self, arg):
        escaped = f'"{arg}"'
        self.assertEqual(escapeShellArgIfNeeded(arg), escaped)

    def assert_does_not_escape(self, arg):
        self.assertEqual(escapeShellArgIfNeeded(arg), arg)

    def test_empty(self):
        self.assert_escapes('')

    def test_spaces(self):
        self.assert_escapes(' ')
        self.assert_escapes('a ')
        self.assert_escapes(' a')
        self.assert_escapes('a b')

    def test_special(self):
        self.assert_escapes('a=b')
        self.assert_escapes('a%b')
        self.assert_escapes('a(b')
        self.assert_escapes('a[b')

    def test_no_escape(self):
        self.assert_does_not_escape('abc')
        self.assert_does_not_escape('a_b')
        self.assert_does_not_escape('-opt')
        self.assert_does_not_escape('--opt')


class TestSetUpGit(unittest.TestCase, config.ConfigErrorsMixin):
    @parameterized.expand([
        ('no_keys', None, None, None, None),
        ('only_private_key', 'key', None, None, None),
        ('private_key_host_key', 'key', 'host', None, None),
        ('private_key_known_hosts', 'key', None, 'hosts', None),
        (
            'no_private_key_host_key',
            None,
            'host',
            None,
            'sshPrivateKey must be provided in order use sshHostKey',
        ),
        (
            'no_private_key_known_hosts',
            None,
            None,
            'hosts',
            'sshPrivateKey must be provided in order use sshKnownHosts',
        ),
        (
            'both_host_key_known_hosts',
            'key',
            'host',
            'hosts',
            'only one of sshKnownHosts and sshHostKey can be provided',
        ),
    ])
    def test_config(self, name, private_key, host_key, known_hosts, config_error):
        if config_error is None:
            check_ssh_config(
                'TestSetUpGit',
                private_key,
                host_key,
                known_hosts,
            )
        else:
            with self.assertRaisesConfigError(config_error):
                check_ssh_config(
                    'TestSetUpGit',
                    private_key,
                    host_key,
                    known_hosts,
                )


class TestParseGitFeatures(GitMixin, unittest.TestCase):
    def setUp(self):
        self.setupGit()

    def test_no_output(self):
        self.parseGitFeatures('')
        self.assertFalse(self.gitInstalled)
        self.assertFalse(self.supportsBranch)
        self.assertFalse(self.supportsSubmoduleForce)
        self.assertFalse(self.supportsSubmoduleCheckout)
        self.assertFalse(self.supportsSshPrivateKeyAsEnvOption)
        self.assertFalse(self.supportsSshPrivateKeyAsConfigOption)
        self.assertFalse(self.supports_lsremote_symref)
        self.assertFalse(self.supports_credential_store)

    def test_git_noversion(self):
        self.parseGitFeatures('git')
        self.assertFalse(self.gitInstalled)
        self.assertFalse(self.supportsBranch)
        self.assertFalse(self.supportsSubmoduleForce)
        self.assertFalse(self.supportsSubmoduleCheckout)
        self.assertFalse(self.supportsSshPrivateKeyAsEnvOption)
        self.assertFalse(self.supportsSshPrivateKeyAsConfigOption)
        self.assertFalse(self.supports_lsremote_symref)
        self.assertFalse(self.supports_credential_store)

    def test_git_zero_version(self):
        self.parseGitFeatures('git version 0.0.0')
        self.assertTrue(self.gitInstalled)
        self.assertFalse(self.supportsBranch)
        self.assertFalse(self.supportsSubmoduleForce)
        self.assertFalse(self.supportsSubmoduleCheckout)
        self.assertFalse(self.supportsSshPrivateKeyAsEnvOption)
        self.assertFalse(self.supportsSshPrivateKeyAsConfigOption)
        self.assertFalse(self.supports_lsremote_symref)
        self.assertFalse(self.supports_credential_store)

    def test_git_2_10_0(self):
        self.parseGitFeatures('git version 2.10.0')
        self.assertTrue(self.gitInstalled)
        self.assertTrue(self.supportsBranch)
        self.assertTrue(self.supportsSubmoduleForce)
        self.assertTrue(self.supportsSubmoduleCheckout)
        self.assertTrue(self.supportsSshPrivateKeyAsEnvOption)
        self.assertTrue(self.supportsSshPrivateKeyAsConfigOption)
        self.assertTrue(self.supports_lsremote_symref)
        self.assertTrue(self.supports_credential_store)


class TestAdjustCommandParamsForSshPrivateKey(GitMixin, unittest.TestCase):
    def test_throws_when_wrapper_not_given(self):
        self.gitInstalled = True

        command = []
        env = {}
        with self.assertRaises(RuntimeError):
            self.setupGit()
            self.adjustCommandParamsForSshPrivateKey(command, env, 'path/to/key')


class TestGetSshKnownHostsContents(unittest.TestCase):
    def test(self):
        key = 'ssh-rsa AAAA<...>WsHQ=='

        expected = '* ssh-rsa AAAA<...>WsHQ=='
        self.assertEqual(expected, getSshKnownHostsContents(key))


class TestensureSshKeyNewline(unittest.TestCase):
    def setUp(self):
        self.sshGoodPrivateKey = """-----BEGIN SSH PRIVATE KEY-----
base64encodedkeydata
-----END SSH PRIVATE KEY-----
"""
        self.sshMissingNewlinePrivateKey = """-----BEGIN SSH PRIVATE KEY-----
base64encodedkeydata
-----END SSH PRIVATE KEY-----"""

    def test_good_key(self):
        """Don't break good keys"""
        self.assertEqual(self.sshGoodPrivateKey, ensureSshKeyNewline(self.sshGoodPrivateKey))

    def test_missing_newline(self):
        """Add missing newline to stripped keys"""
        self.assertEqual(
            self.sshGoodPrivateKey, ensureSshKeyNewline(self.sshMissingNewlinePrivateKey)
        )


class TestScpStyleToUrlSyntax(unittest.TestCase):
    @parameterized.expand([
        ('normal_url', 'ssh://path/to/git', 'ssh://path/to/git'),
        ('unix_path', '/path/to/git', '/path/to/git'),
        ('windows_path', 'C:\\path\\to\\git', 'C:\\path\\to\\git'),
        ('scp_path', 'host:path/to/git', 'ssh://host:23/path/to/git'),
    ])
    def test(self, name, url, expected):
        self.assertEqual(scp_style_to_url_syntax(url, port=23), expected)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import hashlib

import sqlalchemy as sa
from twisted.trial import unittest

from buildbot.test.util import migration
from buildbot.util import sautils


class Migration(migration.MigrateTestMixin, unittest.TestCase):
    def setUp(self):
        return self.setUpMigrateTest()

    def create_tables_thd(self, conn):
        metadata = sa.MetaData()
        metadata.bind = conn

        builders = sautils.Table(
            'builders',
            metadata,
            sa.Column('id', sa.Integer, primary_key=True),
            sa.Column('name', sa.Text, nullable=False),
            sa.Column('description', sa.Text, nullable=True),
            sa.Column('projectid', sa.Integer, nullable=True),
            sa.Column('name_hash', sa.String(40), nullable=False),
        )
        builders.create(bind=conn)

        conn.execute(
            builders.insert(),
            [
                {
                    "id": 3,
                    "name": "foo",
                    "description": "foo_description",
                    "projectid": None,
                    "name_hash": hashlib.sha1(b'foo').hexdigest(),
                }
            ],
        )
        conn.commit()

    def test_update(self):
        def setup_thd(conn):
            self.create_tables_thd(conn)

        def verify_thd(conn):
            metadata = sa.MetaData()
            metadata.bind = conn

            builders = sautils.Table('builders', metadata, autoload_with=conn)
            self.assertIsInstance(builders.c.description_format.type, sa.Text)
            self.assertIsInstance(builders.c.description_html.type, sa.Text)

            q = sa.select(
                builders.c.name,
                builders.c.description_format,
                builders.c.description_html,
            )

            num_rows = 0
            for row in conn.execute(q):
                self.assertIsNone(row.description_format)
                self.assertIsNone(row.description_html)
                num_rows += 1
            self.assertEqual(num_rows, 1)

        return self.do_test_migration('060', '061', setup_thd, verify_thd)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import time
from unittest import mock

from twisted.internet import defer
from twisted.python import log
from twisted.trial import unittest

from buildbot.changes import filter
from buildbot.schedulers import timed
from buildbot.test import fakedb
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util import scheduler
from buildbot.test.util.state import StateTestMixin


class Nightly(scheduler.SchedulerMixin, TestReactorMixin, StateTestMixin, unittest.TestCase):
    OBJECTID = 132
    SCHEDULERID = 32

    # not all timezones are even multiples of 1h from GMT.  This variable
    # holds the number of seconds ahead of the hour for the current timezone.
    # This is then added to the clock before each test is run (to get to 0
    # minutes past the hour) and subtracted before the time offset is reported.
    localtime_offset = time.timezone % 3600

    # Timed scheduler uses the datetime module for some time operations. Windows does not like very
    # small timestamps in these APIs, so tests are adjusted to different times.
    time_offset = 86400 * 10 + localtime_offset

    long_ago_time = 86400

    @defer.inlineCallbacks
    def makeScheduler(self, **kwargs):
        sched = yield self.attachScheduler(
            timed.Nightly(**kwargs), self.OBJECTID, self.SCHEDULERID, overrideBuildsetMethods=True
        )

        yield self.master.db.insert_test_data([
            fakedb.Builder(name=bname) for bname in kwargs.get("builderNames", [])
        ])

        self.reactor.advance(self.time_offset)

        self.addBuildsetCallTimes = []

        def recordTimes(timeList, method):
            def timedMethod(**kw):
                timeList.append(self.reactor.seconds() - self.time_offset)
                return method(**kw)

            return timedMethod

        sched.addBuildsetForSourceStampsWithDefaults = recordTimes(
            self.addBuildsetCallTimes, sched.addBuildsetForSourceStampsWithDefaults
        )
        sched.addBuildsetForChanges = recordTimes(
            self.addBuildsetCallTimes, sched.addBuildsetForChanges
        )

        # see self.assertConsumingChanges
        self.consumingChanges = None

        def startConsumingChanges(**kwargs):
            self.consumingChanges = kwargs
            return defer.succeed(None)

        sched.startConsumingChanges = startConsumingChanges

        return sched

    def mkbs(self, **kwargs):
        # create buildset for expected_buildset in assertBuildset.
        bs = {
            "reason": "The Nightly scheduler named 'test' triggered this build",
            "external_idstring": '',
            "sourcestampsetid": 100,
            "properties": [('scheduler', ('test', 'Scheduler'))],
        }
        bs.update(kwargs)
        return bs

    def mkss(self, **kwargs):
        # create sourcestamp for expected_sourcestamps in assertBuildset.
        ss = {"branch": 'master', "project": '', "repository": '', "sourcestampsetid": 100}
        ss.update(kwargs)
        return ss

    @defer.inlineCallbacks
    def mkch(self, **kwargs):
        # create changeset and insert in database.
        chd = {"branch": 'master', "project": '', "repository": ''}
        chd.update(kwargs)
        ch = self.makeFakeChange(**chd)
        # fakedb.Change requires changeid instead of number
        chd['changeid'] = chd['number']
        del chd['number']
        yield self.master.db.insert_test_data([
            fakedb.Change(**chd),
            fakedb.SourceStamp(id=92),
        ])
        return ch

    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        yield self.setUpScheduler()

    def assertConsumingChanges(self, **kwargs):
        self.assertEqual(self.consumingChanges, kwargs)

    # Tests

    @defer.inlineCallbacks
    def test_constructor_no_reason(self):
        sched = yield self.makeScheduler(name='test', builderNames=['test'], branch='default')
        yield sched.configureService()
        self.assertEqual(sched.reason, "The Nightly scheduler named 'test' triggered this build")

    @defer.inlineCallbacks
    def test_constructor_reason(self):
        sched = yield self.makeScheduler(
            name='test', builderNames=['test'], branch='default', reason="hourly"
        )
        yield sched.configureService()
        self.assertEqual(sched.reason, "hourly")

    @defer.inlineCallbacks
    def test_constructor_change_filter(self):
        sched = yield self.makeScheduler(
            name='test',
            builderNames=['test'],
            branch=None,
            change_filter=filter.ChangeFilter(category_re="fo+o"),
        )
        yield sched.configureService()
        assert sched.change_filter

    @defer.inlineCallbacks
    def test_constructor_month(self):
        sched = yield self.makeScheduler(
            name='test', builderNames=['test'], branch='default', month='1'
        )
        yield sched.configureService()
        self.assertEqual(sched.month, "1")

    @defer.inlineCallbacks
    def test_constructor_priority_none(self):
        sched = yield self.makeScheduler(
            name='test', builderNames=['test'], branch='default', priority=None
        )
        yield sched.configureService()
        self.assertEqual(sched.priority, None)

    @defer.inlineCallbacks
    def test_constructor_priority_int(self):
        sched = yield self.makeScheduler(
            name='test', builderNames=['test'], branch='default', priority=8
        )
        yield sched.configureService()
        self.assertEqual(sched.priority, 8)

    @defer.inlineCallbacks
    def test_constructor_priority_function(self):
        def sched_priority(builderNames, changesByCodebase):
            return 0

        sched = yield self.makeScheduler(
            name='test', builderNames=['test'], branch='default', priority=sched_priority
        )
        yield sched.configureService()
        self.assertEqual(sched.priority, sched_priority)

    @defer.inlineCallbacks
    def test_enabled_callback(self):
        sched = yield self.makeScheduler(name='test', builderNames=['test'], branch='default')
        yield sched.configureService()
        expectedValue = not sched.enabled
        yield sched._enabledCallback(None, {'enabled': not sched.enabled})
        self.assertEqual(sched.enabled, expectedValue)
        expectedValue = not sched.enabled
        yield sched._enabledCallback(None, {'enabled': not sched.enabled})
        self.assertEqual(sched.enabled, expectedValue)

    @defer.inlineCallbacks
    def test_disabled_activate(self):
        sched = yield self.makeScheduler(name='test', builderNames=['test'], branch='default')
        yield sched._enabledCallback(None, {'enabled': not sched.enabled})
        self.assertEqual(sched.enabled, False)
        r = yield sched.activate()
        self.assertEqual(r, None)

    @defer.inlineCallbacks
    def test_disabled_deactivate(self):
        sched = yield self.makeScheduler(name='test', builderNames=['test'], branch='default')
        yield sched._enabledCallback(None, {'enabled': not sched.enabled})
        self.assertEqual(sched.enabled, False)
        r = yield sched.deactivate()
        self.assertEqual(r, None)

    @defer.inlineCallbacks
    def test_disabled_start_build(self):
        sched = yield self.makeScheduler(name='test', builderNames=['test'], branch='default')
        yield sched._enabledCallback(None, {'enabled': not sched.enabled})
        self.assertEqual(sched.enabled, False)
        r = yield sched.startBuild()
        self.assertEqual(r, None)

    # end-to-end tests: let's see the scheduler in action

    @defer.inlineCallbacks
    def test_iterations_simple(self):
        # note that Nightly works in local time, but the TestReactor always
        # starts at midnight UTC, so be careful not to use times that are
        # timezone dependent -- stick to minutes-past-the-half-hour, as some
        # timezones are multiples of 30 minutes off from UTC
        yield self.makeScheduler(
            name='test', builderNames=['test'], branch=None, minute=[10, 20, 21, 40, 50, 51]
        )
        yield self.mkch(number=19)

        # add a change classification
        yield self.master.db.schedulers.classifyChanges(self.SCHEDULERID, {19: True})

        yield self.master.startService()

        # check that the classification has been flushed, since this
        # invocation has not requested onlyIfChanged
        yield self.assert_classifications(self.SCHEDULERID, {})

        self.reactor.advance(0)
        while self.reactor.seconds() < self.time_offset + 30 * 60:
            self.reactor.advance(60)
        self.assertEqual(self.addBuildsetCallTimes, [600, 1200, 1260])
        self.assertEqual(
            self.addBuildsetCalls,
            [
                (
                    'addBuildsetForSourceStampsWithDefaults',
                    {
                        'builderNames': None,
                        'sourcestamps': [{'codebase': ''}],
                        'priority': None,
                        'properties': None,
                        'reason': "The Nightly scheduler named 'test' triggered this build",
                        'waited_for': False,
                    },
                ),
                (
                    'addBuildsetForSourceStampsWithDefaults',
                    {
                        'builderNames': None,
                        'sourcestamps': [{'codebase': ''}],
                        'priority': None,
                        'properties': None,
                        'reason': "The Nightly scheduler named 'test' triggered this build",
                        'waited_for': False,
                    },
                ),
                (
                    'addBuildsetForSourceStampsWithDefaults',
                    {
                        'builderNames': None,
                        'sourcestamps': [{'codebase': ''}],
                        'priority': None,
                        'properties': None,
                        'reason': "The Nightly scheduler named 'test' triggered this build",
                        'waited_for': False,
                    },
                ),
            ],
        )
        yield self.assert_state_by_class('test', 'Nightly', last_build=1260 + self.time_offset)

    @defer.inlineCallbacks
    def test_iterations_simple_with_branch(self):
        # see timezone warning above
        sched = yield self.makeScheduler(
            name='test', builderNames=['test'], branch='master', minute=[5, 35]
        )
        yield self.master.startService()

        yield sched.activate()

        self.reactor.advance(0)
        while self.reactor.seconds() < self.time_offset + 10 * 60:
            self.reactor.advance(60)
        self.assertEqual(self.addBuildsetCallTimes, [300])
        self.assertEqual(
            self.addBuildsetCalls,
            [
                (
                    'addBuildsetForSourceStampsWithDefaults',
                    {
                        'builderNames': None,
                        'sourcestamps': [{'codebase': ''}],
                        'priority': None,
                        'properties': None,
                        'reason': "The Nightly scheduler named 'test' triggered this build",
                        'waited_for': False,
                    },
                )
            ],
        )
        yield self.assert_state_by_class('test', 'Nightly', last_build=300 + self.time_offset)

    @defer.inlineCallbacks
    def do_test_iterations_onlyIfChanged(
        self, changes_at, last_only_if_changed, is_new_scheduler=False, **kwargs
    ):
        fII = mock.Mock(name='fII')
        yield self.makeScheduler(
            name='test',
            builderNames=['test'],
            branch=None,
            minute=[5, 25, 45],
            onlyIfChanged=True,
            fileIsImportant=fII,
            **kwargs,
        )

        if not is_new_scheduler:
            yield self.set_fake_state(self.sched, 'last_build', self.long_ago_time)

        if last_only_if_changed is not None:
            yield self.set_fake_state(self.sched, 'last_only_if_changed', last_only_if_changed)

        return (yield self.do_test_iterations_onlyIfChanged_test(fII, changes_at))

    @defer.inlineCallbacks
    def do_test_iterations_onlyIfChanged_test(self, fII, changes_at):
        yield self.master.startService()

        # check that the scheduler has started to consume changes
        self.assertConsumingChanges(fileIsImportant=fII, change_filter=None, onlyImportant=False)

        # manually run the clock forward through a half-hour, allowing any
        # excitement to take place
        self.reactor.advance(0)  # let it trigger the first build
        while self.reactor.seconds() < self.time_offset + 30 * 60:
            # inject any new changes..
            while changes_at and self.reactor.seconds() >= self.time_offset + changes_at[0][0]:
                _, newchange, important = changes_at.pop(0)
                newchange = yield self.addFakeChange(newchange)
                yield self.sched.gotChange(newchange, important).addErrback(log.err)
            # and advance the clock by a minute
            self.reactor.advance(60)

    @defer.inlineCallbacks
    def test_iterations_onlyIfChanged_no_changes_new_scheduler(self):
        yield self.do_test_iterations_onlyIfChanged(
            [], last_only_if_changed=None, is_new_scheduler=True
        )
        self.assertEqual(
            self.addBuildsetCalls,
            [
                (
                    'addBuildsetForSourceStampsWithDefaults',
                    {
                        'builderNames': None,
                        'priority': None,
                        'properties': None,
                        'reason': "The Nightly scheduler named 'test' triggered this build",
                        'sourcestamps': [{'codebase': ''}],
                        'waited_for': False,
                    },
                )
            ],
        )
        yield self.assert_state_by_class('test', 'Nightly', last_build=1500 + self.time_offset)
        yield self.sched.deactivate()

    @defer.inlineCallbacks
    def test_iterations_onlyIfChanged_no_changes_existing_scheduler(self):
        yield self.do_test_iterations_onlyIfChanged([], last_only_if_changed=True)
        self.assertEqual(self.addBuildsetCalls, [])
        yield self.assert_state_by_class('test', 'Nightly', last_build=1500 + self.time_offset)
        yield self.sched.deactivate()

    @defer.inlineCallbacks
    def test_iterations_onlyIfChanged_no_changes_existing_scheduler_setting_changed(self):
        # When onlyIfChanged==False, builds are run every time on the time set
        # (changes or no changes). Changes are being recognized but do not have any effect on
        # starting builds.
        # It might happen that onlyIfChanged was False, then change happened, then setting was
        # changed to onlyIfChanged==True.
        # Because onlyIfChanged was False possibly important change will be missed.
        # Therefore the first build should start immediately.

        yield self.do_test_iterations_onlyIfChanged([], last_only_if_changed=False)
        self.assertEqual(
            self.addBuildsetCalls,
            [
                (
                    'addBuildsetForSourceStampsWithDefaults',
                    {
                        'builderNames': None,
                        'priority': None,
                        'properties': None,
                        'reason': "The Nightly scheduler named 'test' triggered this build",
                        'sourcestamps': [{'codebase': ''}],
                        'waited_for': False,
                    },
                )
            ],
        )
        yield self.assert_state_by_class('test', 'Nightly', last_build=1500 + self.time_offset)
        yield self.sched.deactivate()

    @defer.inlineCallbacks
    def test_iterations_onlyIfChanged_no_changes_existing_scheduler_update_to_v3_5_0(self):
        # v3.4.0 have not had a variable last_only_if_changed yet therefore this case is tested
        # separately
        yield self.do_test_iterations_onlyIfChanged([], last_only_if_changed=None)
        self.assertEqual(self.addBuildsetCallTimes, [])
        self.assertEqual(self.addBuildsetCalls, [])
        yield self.assert_state_by_class('test', 'Nightly', last_build=1500 + self.time_offset)
        yield self.sched.deactivate()

    @defer.inlineCallbacks
    def test_iterations_onlyIfChanged_no_changes_force_at(self):
        yield self.do_test_iterations_onlyIfChanged(
            [], last_only_if_changed=True, force_at_minute=[23, 25, 27]
        )

        self.assertEqual(self.addBuildsetCallTimes, [1500])
        self.assertEqual(
            self.addBuildsetCalls,
            [
                (
                    'addBuildsetForSourceStampsWithDefaults',
                    {
                        'builderNames': None,
                        'priority': None,
                        'properties': None,
                        'reason': "The Nightly scheduler named 'test' triggered this build",
                        'sourcestamps': [{'codebase': ''}],
                        'waited_for': False,
                    },
                )
            ],
        )
        yield self.assert_state_by_class('test', 'Nightly', last_build=1500 + self.time_offset)
        yield self.sched.deactivate()

    @defer.inlineCallbacks
    def test_iterations_onlyIfChanged_unimp_changes_calls_for_new_scheduler(self):
        yield self.do_test_iterations_onlyIfChanged(
            [
                (60, self.makeFakeChange(number=500, branch=None), False),
                (600, self.makeFakeChange(number=501, branch=None), False),
            ],
            last_only_if_changed=None,
            is_new_scheduler=True,
        )

        self.assertEqual(
            self.addBuildsetCalls,
            [
                (
                    'addBuildsetForChanges',
                    {
                        'builderNames': None,
                        'changeids': [500],
                        'external_idstring': None,
                        'priority': None,
                        'properties': None,
                        'reason': "The Nightly scheduler named 'test' triggered this build",
                        'waited_for': False,
                    },
                )
            ],
        )

        yield self.assert_state_by_class('test', 'Nightly', last_build=1500 + self.time_offset)
        yield self.sched.deactivate()

    @defer.inlineCallbacks
    def test_iterations_onlyIfChanged_unimp_changes_existing_sched_changed_only_if_changed(self):
        yield self.do_test_iterations_onlyIfChanged(
            [
                (60, self.makeFakeChange(number=500, branch=None), False),
                (600, self.makeFakeChange(number=501, branch=None), False),
            ],
            last_only_if_changed=False,
        )

        self.assertEqual(
            self.addBuildsetCalls,
            [
                (
                    'addBuildsetForSourceStampsWithDefaults',
                    {
                        'builderNames': None,
                        'priority': None,
                        'properties': None,
                        'reason': "The Nightly scheduler named 'test' triggered this build",
                        'sourcestamps': [{'codebase': ''}],
                        'waited_for': False,
                    },
                )
            ],
        )

        yield self.assert_state_by_class('test', 'Nightly', last_build=1500 + self.time_offset)
        yield self.sched.deactivate()

    @defer.inlineCallbacks
    def test_iterations_onlyIfChanged_unimp_changes_existing_sched_same_only_if_changed(self):
        yield self.do_test_iterations_onlyIfChanged(
            [
                (60, self.makeFakeChange(number=500, branch=None), False),
                (600, self.makeFakeChange(number=501, branch=None), False),
            ],
            last_only_if_changed=True,
        )

        self.assertEqual(self.addBuildsetCalls, [])

        yield self.assert_state_by_class('test', 'Nightly', last_build=1500 + self.time_offset)
        yield self.sched.deactivate()

    @defer.inlineCallbacks
    def test_iterations_onlyIfChanged_changes_existing_scheduler_update_to_v3_5_0(self):
        # v3.4.0 have not had a variable last_only_if_changed yet therefore this case is tested
        # separately
        yield self.do_test_iterations_onlyIfChanged(
            [
                (120, self.makeFakeChange(number=500, branch=None), False),
                (1200, self.makeFakeChange(number=501, branch=None), True),
                (1201, self.makeFakeChange(number=502, branch=None), False),
            ],
            last_only_if_changed=None,
        )

        self.assertEqual(self.addBuildsetCallTimes, [1500])
        self.assertEqual(
            self.addBuildsetCalls,
            [
                (
                    'addBuildsetForChanges',
                    {
                        'waited_for': False,
                        'reason': "The Nightly scheduler named 'test' triggered this build",
                        'external_idstring': None,
                        'changeids': [500, 501, 502],
                        'priority': None,
                        'properties': None,
                        'builderNames': None,
                    },
                ),
            ],
        )
        yield self.assert_state_by_class('test', 'Nightly', last_build=1500 + self.time_offset)
        yield self.sched.deactivate()

    @defer.inlineCallbacks
    def test_iterations_onlyIfChanged_unimp_changes_force_at(self):
        yield self.do_test_iterations_onlyIfChanged(
            [
                (60, self.makeFakeChange(number=500, branch=None), False),
                (600, self.makeFakeChange(number=501, branch=None), False),
            ],
            last_only_if_changed=True,
            force_at_minute=[23, 25, 27],
        )

        self.assertEqual(self.addBuildsetCallTimes, [1500])
        self.assertEqual(
            self.addBuildsetCalls,
            [
                (
                    'addBuildsetForChanges',
                    {
                        'builderNames': None,
                        'changeids': [500, 501],
                        'external_idstring': None,
                        'priority': None,
                        'properties': None,
                        'reason': "The Nightly scheduler named 'test' triggered this build",
                        'waited_for': False,
                    },
                )
            ],
        )
        yield self.assert_state_by_class('test', 'Nightly', last_build=1500 + self.time_offset)
        yield self.sched.deactivate()

    @defer.inlineCallbacks
    def test_iterations_onlyIfChanged_off_branch_changes(self):
        yield self.do_test_iterations_onlyIfChanged(
            [
                (60, self.makeFakeChange(number=500, branch='testing'), True),
                (1700, self.makeFakeChange(number=501, branch='staging'), True),
            ],
            last_only_if_changed=True,
        )

        self.assertEqual(self.addBuildsetCalls, [])
        yield self.assert_state_by_class('test', 'Nightly', last_build=1500 + self.time_offset)
        yield self.sched.deactivate()

    @defer.inlineCallbacks
    def test_iterations_onlyIfChanged_mixed_changes(self):
        yield self.do_test_iterations_onlyIfChanged(
            [
                (120, self.makeFakeChange(number=500, branch=None), False),
                (130, self.makeFakeChange(number=501, branch='offbranch'), True),
                (1200, self.makeFakeChange(number=502, branch=None), True),
                (1201, self.makeFakeChange(number=503, branch=None), False),
                (1202, self.makeFakeChange(number=504, branch='offbranch'), True),
            ],
            last_only_if_changed=True,
        )

        # note that the changeid list includes the unimportant changes, but not the
        # off-branch changes, and note that no build took place at 300s, as no important
        # changes had yet arrived
        self.assertEqual(self.addBuildsetCallTimes, [1500])
        self.assertEqual(
            self.addBuildsetCalls,
            [
                (
                    'addBuildsetForChanges',
                    {
                        'builderNames': None,
                        'changeids': [500, 502, 503],
                        'external_idstring': None,
                        'priority': None,
                        'properties': None,
                        'reason': "The Nightly scheduler named 'test' triggered this build",
                        'waited_for': False,
                    },
                )
            ],
        )
        yield self.assert_state_by_class('test', 'Nightly', last_build=1500 + self.time_offset)
        yield self.sched.deactivate()

    @defer.inlineCallbacks
    def test_iterations_onlyIfChanged_createAbsoluteSourceStamps_oneChanged(self):
        # Test createAbsoluteSourceStamps=True when only one codebase has
        # changed
        yield self.do_test_iterations_onlyIfChanged(
            [
                (120, self.makeFakeChange(number=500, codebase='a', revision='2345:bcd'), True),
            ],
            codebases={
                'a': {'repository': "", 'branch': 'master'},
                'b': {'repository': "", 'branch': 'master'},
            },
            createAbsoluteSourceStamps=True,
            last_only_if_changed=True,
        )
        yield self.assert_state_by_class('test', 'Nightly', last_build=1500 + self.time_offset)
        # addBuildsetForChanges calls getCodebase, so this isn't too
        # interesting
        self.assertEqual(self.addBuildsetCallTimes, [300])
        self.assertEqual(
            self.addBuildsetCalls,
            [
                (
                    'addBuildsetForChanges',
                    {
                        'builderNames': None,
                        'changeids': [500],
                        'external_idstring': None,
                        'priority': None,
                        'properties': None,
                        'reason': "The Nightly scheduler named 'test' triggered this build",
                        'waited_for': False,
                    },
                )
            ],
        )
        yield self.assert_state_by_class(
            'test',
            'Nightly',
            lastCodebases={
                'a': {"revision": '2345:bcd', "branch": None, "repository": '', "lastChange": 500}
            },
        )
        yield self.sched.deactivate()

    @defer.inlineCallbacks
    def test_iterations_onlyIfChanged_createAbsoluteSourceStamps_oneChanged_loadOther(self):
        # Test createAbsoluteSourceStamps=True when only one codebase has changed,
        # but the other was previously changed
        fII = mock.Mock(name='fII')
        yield self.makeScheduler(
            name='test',
            builderNames=['test'],
            branch=None,
            minute=[5, 25, 45],
            onlyIfChanged=True,
            fileIsImportant=fII,
            codebases={
                'a': {'repository': "", 'branch': 'master'},
                'b': {'repository': "", 'branch': 'master'},
            },
            createAbsoluteSourceStamps=True,
        )

        yield self.set_fake_state(self.sched, 'last_only_if_changed', True)
        yield self.set_fake_state(
            self.sched,
            'lastCodebases',
            {
                'b': {
                    'branch': 'master',
                    'repository': 'B',
                    'revision': '1234:abc',
                    'lastChange': 499,
                }
            },
        )

        yield self.do_test_iterations_onlyIfChanged_test(
            fII,
            [
                (120, self.makeFakeChange(number=500, codebase='a', revision='2345:bcd'), True),
            ],
        )

        yield self.assert_state_by_class('test', 'Nightly', last_build=1500 + self.time_offset)
        # addBuildsetForChanges calls getCodebase, so this isn't too
        # interesting
        self.assertEqual(self.addBuildsetCallTimes, [300])
        self.assertEqual(
            self.addBuildsetCalls,
            [
                (
                    'addBuildsetForChanges',
                    {
                        'builderNames': None,
                        'changeids': [500],
                        'external_idstring': None,
                        'priority': None,
                        'properties': None,
                        'reason': "The Nightly scheduler named 'test' triggered this build",
                        'waited_for': False,
                    },
                )
            ],
        )
        yield self.assert_state_by_class(
            'test',
            'Nightly',
            lastCodebases={
                'a': {"revision": '2345:bcd', "branch": None, "repository": '', "lastChange": 500},
                'b': {
                    "revision": '1234:abc',
                    "branch": "master",
                    "repository": 'B',
                    "lastChange": 499,
                },
            },
        )
        yield self.sched.deactivate()

    @defer.inlineCallbacks
    def test_iterations_onlyIfChanged_createAbsoluteSourceStamps_bothChanged(self):
        # Test createAbsoluteSourceStamps=True when both codebases have changed
        yield self.do_test_iterations_onlyIfChanged(
            [
                (120, self.makeFakeChange(number=500, codebase='a', revision='2345:bcd'), True),
                (122, self.makeFakeChange(number=501, codebase='b', revision='1234:abc'), True),
            ],
            codebases={
                'a': {'repository': "", 'branch': 'master'},
                'b': {'repository': "", 'branch': 'master'},
            },
            last_only_if_changed=None,
            createAbsoluteSourceStamps=True,
        )

        yield self.assert_state_by_class('test', 'Nightly', last_build=1500 + self.time_offset)
        # addBuildsetForChanges calls getCodebase, so this isn't too
        # interesting
        self.assertEqual(self.addBuildsetCallTimes, [300])
        self.assertEqual(
            self.addBuildsetCalls,
            [
                (
                    'addBuildsetForChanges',
                    {
                        'builderNames': None,
                        'changeids': [500, 501],
                        'external_idstring': None,
                        'priority': None,
                        'properties': None,
                        'reason': "The Nightly scheduler named 'test' triggered this build",
                        'waited_for': False,
                    },
                )
            ],
        )
        yield self.assert_state_by_class(
            'test',
            'Nightly',
            lastCodebases={
                'a': {"revision": '2345:bcd', "branch": None, "repository": '', "lastChange": 500},
                'b': {"revision": '1234:abc', "branch": None, "repository": '', "lastChange": 501},
            },
        )
        yield self.sched.deactivate()



#!/usr/local/bin/python
#
# BitKeeper hook script.
#
# svn_buildbot.py was used as a base for this file, if you find any bugs or
# errors please email me.
#
# Amar Takhar <amar@ntp.org>


import subprocess
import sys

from twisted.cred import credentials
from twisted.internet import reactor
from twisted.python import usage
from twisted.spread import pb

'''
/path/to/bk_buildbot.py --repository "$REPOS" --revision "$REV" --branch \
"<branch>" --bbserver localhost --bbport 9989
'''


# We have hackish "-d" handling here rather than in the Options
# subclass below because a common error will be to not have twisted in
# PYTHONPATH; we want to be able to print that error to the log if
# debug mode is on, so we set it up before the imports.

DEBUG = None

if '-d' in sys.argv:
    i = sys.argv.index('-d')
    DEBUG = sys.argv[i + 1]
    del sys.argv[i]
    del sys.argv[i]

if DEBUG:
    f = open(DEBUG, 'a')
    sys.stderr = f
    sys.stdout = f


class Options(usage.Options):
    optParameters = [
        ['repository', 'r', None, "The repository that was changed."],
        ['revision', 'v', None, "The revision that we want to examine (default: latest)"],
        ['branch', 'b', None, "Name of the branch to insert into the branch field. (REQUIRED)"],
        ['category', 'c', None, "Schedular category."],
        ['bbserver', 's', 'localhost', "The hostname of the server that buildbot is running on"],
        ['bbport', 'p', 8007, "The port that buildbot is listening on"],
    ]
    optFlags = [
        ['dryrun', 'n', "Do not actually send changes"],
    ]

    def __init__(self):
        usage.Options.__init__(self)

    def postOptions(self):
        if self['repository'] is None:
            raise usage.error("You must pass --repository")


class ChangeSender:
    def getChanges(self, opts):
        """Generate and stash a list of Change dictionaries, ready to be sent
        to the buildmaster's PBChangeSource."""

        # first we extract information about the files that were changed
        repo = opts['repository']
        print("Repo:", repo)
        rev_arg = ''
        if opts['revision']:
            rev_arg = '-r"{}"'.format(opts['revision'])
        changed = subprocess.check_output(
            f"bk changes -v {rev_arg} -d':GFILE:\\n' '{repo}'", shell=True
        )
        changed = changed.decode(sys.stdout.encoding)
        changed = changed.split('\n')

        # Remove the first line, it's an info message you can't remove
        # (annoying)
        del changed[0]

        change_info = subprocess.check_output(
            f"bk changes {rev_arg} -d':USER:\\n$each(:C:){{(:C:)\\n}}' '{repo}'", shell=True
        )
        change_info = change_info.decode(sys.stdout.encoding)
        change_info = change_info.split('\n')

        # Remove the first line, it's an info message you can't remove
        # (annoying)
        del change_info[0]

        who = change_info.pop(0)
        branch = opts['branch']
        message = '\n'.join(change_info)
        revision = opts.get('revision')

        changes = {
            'who': who,
            'branch': branch,
            'files': changed,
            'comments': message,
            'revision': revision,
        }

        if opts.get('category'):
            changes['category'] = opts.get('category')

        return changes

    def sendChanges(self, opts, changes):
        pbcf = pb.PBClientFactory()
        reactor.connectTCP(opts['bbserver'], int(opts['bbport']), pbcf)
        d = pbcf.login(credentials.UsernamePassword('change', 'changepw'))
        d.addCallback(self.sendAllChanges, changes)
        return d

    def sendAllChanges(self, remote, changes):
        dl = remote.callRemote('addChange', changes)
        return dl

    def run(self):
        opts = Options()
        try:
            opts.parseOptions()
            if not opts['branch']:
                print("You must supply a branch with -b or --branch.")
                sys.exit(1)

        except usage.error as ue:
            print(opts)
            print(f"{sys.argv[0]}: {ue}")
            sys.exit()

        changes = self.getChanges(opts)
        if opts['dryrun']:
            for k in changes.keys():
                print(f"[{k:>10}]: {changes[k]}")
            print("*NOT* sending any changes")
            return

        d = self.sendChanges(opts, changes)

        def quit(*why):
            print("quitting! because", why)
            reactor.stop()

        @d.addErrback
        def failed(f):
            print(f"FAILURE: {f}")
            reactor.stop()

        d.addCallback(quit, "SUCCESS")
        reactor.callLater(60, quit, "TIMEOUT")

        reactor.run()


if __name__ == '__main__':
    s = ChangeSender()
    s.run()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
from __future__ import annotations

import re
from typing import TYPE_CHECKING

from twisted.logger import Logger

if TYPE_CHECKING:
    from typing_extensions import TypeAlias

log = Logger()


LineInfo: TypeAlias = "tuple[str, list[int], list[float]]"


class LineBoundaryFinder:
    __slots__ = ['max_line_length', 'newline_re', 'partial_line', 'time', 'warned']

    def __init__(self, max_line_length: int, newline_re: str) -> None:
        # split at reasonable line length.
        # too big lines will fill master's memory, and slow down the UI too much.
        self.max_line_length = max_line_length
        self.newline_re = re.compile(newline_re)
        self.partial_line = ""
        self.warned = False
        self.time: float | None = None

    def append(self, text: str, time: float) -> LineInfo | None:
        # returns a tuple containing three elements:
        # - text: string containing one or more lines
        # - lf_positions: newline position in returned string
        # - line times: times when first line symbol was received
        had_partial_line = False
        time_partial_line = None
        if self.partial_line:
            had_partial_line = True
            text = self.partial_line + text
            assert self.time is not None
            time_partial_line = self.time

        text = self.newline_re.sub('\n', text)

        lf_positions = self.get_lf_positions(text)

        # lines with appropriate number of symbols and their separators \n
        ret_lines: list[str] = []
        ret_indexes: list[int] = []  # ret_indexes is a list of '\n' symbols
        ret_text_length = -1
        ret_line_count = 0

        first_position = 0
        for position in lf_positions:
            # finds too long lines and splits them, each element in ret_lines will be a line of
            # appropriate length
            while position - first_position >= self.max_line_length:
                line = text[first_position : self.max_line_length - 1] + '\n'
                ret_lines.append(line)
                ret_line_count += 1
                ret_text_length = ret_text_length + len(line)
                ret_indexes.append(ret_text_length)
                first_position = first_position + self.max_line_length

            line = text[first_position : (position + 1)]
            ret_lines.append(line)
            ret_line_count += 1
            ret_text_length = ret_text_length + len(line)
            ret_indexes.append(ret_text_length)
            first_position = position + 1

        position = len(text)
        while position - first_position >= self.max_line_length:
            line = text[first_position : self.max_line_length - 1] + '\n'
            ret_lines.append(line)
            ret_text_length = ret_text_length + len(line)
            ret_indexes.append(ret_text_length)
            first_position = first_position + self.max_line_length - 1

        if had_partial_line:
            times = []
            if ret_line_count > 1:
                times = [time] * (ret_line_count - 1)

            assert time_partial_line is not None
            line_times = [time_partial_line, *times]
        else:
            line_times = ret_line_count * [time]

        ret_text = "".join(ret_lines)

        if ret_text != '' or not had_partial_line:
            self.time = time

        self.partial_line = text[first_position:position]

        if ret_text == '':
            return None

        return (ret_text, ret_indexes, line_times)

    def get_lf_positions(self, text: str) -> list[int]:
        lf_position = 0
        lf_positions: list[int] = []
        while lf_position != -1:
            lf_position = text.find('\n', lf_position)
            if lf_position < 0:
                break
            lf_positions.append(lf_position)
            lf_position = lf_position + 1
        return lf_positions

    def flush(self) -> LineInfo | None:
        if self.partial_line != "":
            assert self.time is not None
            return self.append('\n', self.time)
        return None



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from twisted.internet import defer

from buildbot.mq import base
from buildbot.test.util import validation
from buildbot.util import deferwaiter
from buildbot.util import service
from buildbot.util import tuplematch
from buildbot.util.twisted import async_to_deferred


class FakeMQConnector(service.AsyncMultiService, base.MQBase):
    # a fake connector that doesn't actually bridge messages from production to
    # consumption, and thus doesn't do any topic handling or persistence

    # note that this *does* verify all messages sent and received, unless this
    # is set to false:
    verifyMessages = True

    def __init__(self, testcase):
        super().__init__()
        self.testcase = testcase
        self.setup_called = False
        self.productions = []
        self.qrefs = []
        self._deferwaiter = deferwaiter.DeferWaiter()

    @async_to_deferred
    async def stopService(self) -> None:
        await self._deferwaiter.wait()
        await super().stopService()

    def setup(self):
        self.setup_called = True
        return defer.succeed(None)

    def produce(self, routingKey, data):
        self.testcase.assertIsInstance(routingKey, tuple)

        # XXX this is incompatible with the new scheme of sending multiple messages,
        # since the message type is no longer encoded by the first element of the
        # routing key
        #        if self.verifyMessages:
        #            validation.verifyMessage(self.testcase, routingKey, data)

        if any(not isinstance(k, str) for k in routingKey):
            raise AssertionError(f"{routingKey} is not all str")
        self.productions.append((routingKey, data))
        # note - no consumers are called: IT'S A FAKE

    def callConsumer(self, routingKey, msg):
        if self.verifyMessages:
            validation.verifyMessage(self.testcase, routingKey, msg)
        matched = False
        for q in self.qrefs:
            if tuplematch.matchTuple(routingKey, q.filter):
                matched = True
                self._deferwaiter.add(q.callback(routingKey, msg))
        if not matched:
            raise AssertionError("no consumer found")

    def startConsuming(self, callback, filter, persistent_name=None):
        if any(not isinstance(k, str) and k is not None for k in filter):
            raise AssertionError(f"{filter} is not a filter")
        qref = FakeQueueRef()
        qref.qrefs = self.qrefs
        qref.callback = callback
        qref.filter = filter
        qref.persistent_name = persistent_name
        self.qrefs.append(qref)
        return defer.succeed(qref)

    def clearProductions(self):
        "Clear out the cached productions"
        self.productions = []

    def assertProductions(self, exp, orderMatters=True):
        """Assert that the given messages have been produced, then flush the
        list of produced messages.

        If C{orderMatters} is false, then the messages are sorted first; use
        this in cases where the messages must all be produced, but the order is
        not specified.
        """
        if orderMatters:
            self.testcase.assertEqual(self.productions, exp)
        else:
            self.testcase.assertEqual(sorted(self.productions), sorted(exp))
        self.productions = []

    @async_to_deferred
    async def wait_consumed(self) -> None:
        # waits until all messages have been consumed
        await self._deferwaiter.wait()


class FakeQueueRef:
    def stopConsuming(self):
        if self in self.qrefs:
            self.qrefs.remove(self)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

import dataclasses
from typing import TYPE_CHECKING

import sqlalchemy as sa
from twisted.python import deprecate
from twisted.python import versions

from buildbot.db import base
from buildbot.util import epoch2datetime
from buildbot.util.sautils import hash_columns
from buildbot.warnings import warn_deprecated

if TYPE_CHECKING:
    import datetime

    from twisted.internet import defer


@dataclasses.dataclass
class MasterModel:
    id: int
    name: str
    active: bool
    last_active: datetime.datetime

    # For backward compatibility
    def __getitem__(self, key: str):
        warn_deprecated(
            '4.1.0',
            (
                'MastersConnectorComponent '
                'getMaster, and getMasters '
                'no longer return Master as dictionnaries. '
                'Usage of [] accessor is deprecated: please access the member directly'
            ),
        )

        if hasattr(self, key):
            return getattr(self, key)

        raise KeyError(key)


@deprecate.deprecated(versions.Version("buildbot", 4, 1, 0), MasterModel)
class MasterDict(dict):
    pass


class MastersConnectorComponent(base.DBConnectorComponent):
    data2db = {"masterid": "id", "link": "id"}

    def findMasterId(self, name: str) -> defer.Deferred[int]:
        tbl = self.db.model.masters
        name_hash = hash_columns(name)
        return self.findSomethingId(
            tbl=tbl,
            whereclause=(tbl.c.name_hash == name_hash),
            insert_values={
                "name": name,
                "name_hash": name_hash,
                "active": 0,  # initially inactive
                "last_active": int(self.master.reactor.seconds()),
            },
        )

    def setMasterState(self, masterid: int, active: bool) -> defer.Deferred[bool]:
        def thd(conn) -> bool:
            tbl = self.db.model.masters
            whereclause = tbl.c.id == masterid

            # get the old state
            r = conn.execute(sa.select(tbl.c.active).where(whereclause))
            rows = r.fetchall()
            r.close()
            if not rows:
                return False  # can't change a row that doesn't exist..
            was_active = bool(rows[0].active)

            if not active:
                # if we're marking inactive, then delete any links to this
                # master
                sch_mst_tbl = self.db.model.scheduler_masters
                q = sch_mst_tbl.delete().where(sch_mst_tbl.c.masterid == masterid)
                conn.execute(q)
                conn.commit()

            # set the state (unconditionally, just to be safe)
            q = tbl.update().where(whereclause)
            q = q.values(active=1 if active else 0)
            if active:
                q = q.values(last_active=int(self.master.reactor.seconds()))
            conn.execute(q)
            conn.commit()

            # return True if there was a change in state
            return was_active != bool(active)

        return self.db.pool.do(thd)

    def getMaster(self, masterid: int) -> defer.Deferred[MasterModel | None]:
        def thd(conn) -> MasterModel | None:
            tbl = self.db.model.masters
            res = conn.execute(tbl.select().where(tbl.c.id == masterid))
            row = res.fetchone()

            rv = None
            if row:
                rv = self._model_from_row(row)
            res.close()
            return rv

        return self.db.pool.do(thd)

    def getMasters(self) -> defer.Deferred[list[MasterModel]]:
        def thd(conn) -> list[MasterModel]:
            tbl = self.db.model.masters
            return [self._model_from_row(row) for row in conn.execute(tbl.select()).fetchall()]

        return self.db.pool.do(thd)

    def setAllMastersActiveLongTimeAgo(self) -> defer.Deferred[None]:
        def thd(conn) -> None:
            tbl = self.db.model.masters
            q = tbl.update().values(active=1, last_active=0)
            conn.execute(q)

        return self.db.pool.do(thd)

    def _model_from_row(self, row):
        return MasterModel(
            id=row.id,
            name=row.name,
            active=bool(row.active),
            last_active=epoch2datetime(row.last_active),
        )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
from __future__ import annotations

from parameterized import parameterized
from twisted.trial import unittest

from buildbot_worker import util


class remove_userpassword(unittest.TestCase):
    def assertUrl(self, real_url: str, expected_url: str) -> None:
        new_url = util.remove_userpassword(real_url)
        self.assertEqual(expected_url, new_url)

    def test_url_with_no_user_and_password(self) -> None:
        self.assertUrl('http://myurl.com/myrepo', 'http://myurl.com/myrepo')

    def test_url_with_user_and_password(self) -> None:
        self.assertUrl('http://myuser:mypass@myurl.com/myrepo', 'http://myurl.com/myrepo')

    def test_another_url_with_no_user_and_password(self) -> None:
        self.assertUrl('http://myurl2.com/myrepo2', 'http://myurl2.com/myrepo2')

    def test_another_url_with_user_and_password(self) -> None:
        self.assertUrl('http://myuser2:mypass2@myurl2.com/myrepo2', 'http://myurl2.com/myrepo2')

    def test_with_different_protocol_without_user_and_password(self) -> None:
        self.assertUrl('ssh://myurl3.com/myrepo3', 'ssh://myurl3.com/myrepo3')

    def test_with_different_protocol_with_user_and_password(self) -> None:
        self.assertUrl('ssh://myuser3:mypass3@myurl3.com/myrepo3', 'ssh://myurl3.com/myrepo3')

    def test_file_path(self) -> None:
        self.assertUrl('/home/me/repos/my-repo', '/home/me/repos/my-repo')

    def test_file_path_with_at_sign(self) -> None:
        self.assertUrl('/var/repos/speci@l', '/var/repos/speci@l')

    def test_win32file_path(self) -> None:
        self.assertUrl('c:\\repos\\my-repo', 'c:\\repos\\my-repo')


class TestObfuscated(unittest.TestCase):
    def testSimple(self) -> None:
        c = util.Obfuscated('real', '****')
        self.assertEqual(str(c), '****')
        self.assertEqual(repr(c), "'****'")

    def testObfuscatedCommand(self) -> None:
        cmd = ['echo', util.Obfuscated('password', '*******')]
        cmd_bytes = [b'echo', util.Obfuscated(b'password', b'*******')]
        cmd_unicode = ['echo', util.Obfuscated('password', 'привет')]

        self.assertEqual(['echo', 'password'], util.Obfuscated.get_real(cmd))
        self.assertEqual(['echo', '*******'], util.Obfuscated.get_fake(cmd))
        self.assertEqual([b'echo', b'password'], util.Obfuscated.get_real(cmd_bytes))
        self.assertEqual([b'echo', b'*******'], util.Obfuscated.get_fake(cmd_bytes))
        self.assertEqual(['echo', 'password'], util.Obfuscated.get_real(cmd_unicode))
        self.assertEqual(['echo', 'привет'], util.Obfuscated.get_fake(cmd_unicode))

    def testObfuscatedNonString(self) -> None:
        cmd = ['echo', 1]
        cmd_bytes = [b'echo', 2]
        cmd_unicode = ['привет', 3]
        self.assertEqual(['echo', '1'], util.Obfuscated.get_real(cmd))
        self.assertEqual([b'echo', '2'], util.Obfuscated.get_fake(cmd_bytes))
        self.assertEqual(['привет', '3'], util.Obfuscated.get_fake(cmd_unicode))

    def testObfuscatedNonList(self) -> None:
        cmd = 1
        self.assertEqual(1, util.Obfuscated.get_real(cmd))
        self.assertEqual(1, util.Obfuscated.get_fake(cmd))


class TestRewrap(unittest.TestCase):
    def test_main(self) -> None:
        tests = [
            ("", "", None),
            ("\n", "\n", None),
            ("\n  ", "\n", None),
            ("  \n", "\n", None),
            ("  \n  ", "\n", None),
            (
                """
                multiline
                with
                indent
                """,
                "\nmultiline with indent",
                None,
            ),
            (
                """\
                multiline
                with
                indent

                """,
                "multiline with indent\n",
                None,
            ),
            (
                """\
                 multiline
                 with
                 indent

                 """,
                "multiline with indent\n",
                None,
            ),
            (
                """\
                multiline
                with
                indent

                  and
                   formatting
                """,
                "multiline with indent\n  and\n   formatting\n",
                None,
            ),
            (
                """\
                multiline
                with
                indent
                and wrapping

                  and
                   formatting
                """,
                "multiline with\nindent and\nwrapping\n  and\n   formatting\n",
                15,
            ),
        ]

        for text, expected, width in tests:
            self.assertEqual(util.rewrap(text, width=width), expected)


class TestTwistedConnectionStringToWsUrl(unittest.TestCase):
    @parameterized.expand([
        ('empty', ''),
        ('tcp_no_host', 'tcp:'),
        ('tls_no_host', 'tls:'),
    ])
    def test_exception(self, name: str, description: str) -> None:
        with self.assertRaises(ValueError):
            util.twisted_connection_string_to_ws_url(description)

    @parameterized.expand([
        ('tcp_host', 'tcp:abc', 'ws://abc:80'),
        ('tcp_host_port', 'tcp:abc:123', 'ws://abc:123'),
        ('tcp_host_kw_port', 'tcp:host=abc:123', 'ws://abc:123'),
        ('tcp_host_port_kw', 'tcp:abc:port=123', 'ws://abc:123'),
        ('tcp_host_kw_port_kw', 'tcp:host=abc:port=123', 'ws://abc:123'),
        ('tls_host_port', 'tls:host=abc:port=123', 'ws://abc:123'),
    ])
    def test_converts(self, name: str, description: str, expected: str) -> None:
        ws_connection = util.twisted_connection_string_to_ws_url(description)
        self.assertEqual(ws_connection, expected)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import os
import shutil
import sys
from subprocess import call
from subprocess import check_call
from textwrap import dedent

from twisted.trial import unittest


class BuildbotWWWPkg(unittest.TestCase):
    pkgName = "buildbot_www"
    pkgPaths = ["www", "base"]
    epName = "base"

    loadTestScript = dedent(r"""
        from importlib.metadata import entry_points
        import re
        apps = {}

        eps = entry_points()
        entry_point = "buildbot.www"
        if hasattr(eps, "select"):
            entry_point_group = eps.select(group=entry_point)
        else:
            entry_point_group = eps.get(entry_point, [])

        for ep in entry_point_group:
            apps[ep.name] = ep.load()

        print(apps["%(epName)s"])
        print(apps["%(epName)s"].resource.listNames())
        expected_file = "scripts.js"
        if "%(epName)s" == "base":
            expected_file = "index.html"
        assert(expected_file in apps["%(epName)s"].resource.listNames())
        assert(re.match(r'\d+\.\d+\.\d+', apps["%(epName)s"].version) is not None)
        assert(apps["%(epName)s"].description is not None)
        """)

    @property
    def path(self):
        return os.path.abspath(os.path.join(os.path.dirname(__file__), "..", *self.pkgPaths))

    def rmtree(self, d):
        if os.path.isdir(d):
            shutil.rmtree(d)

    def setUp(self):
        call("pip uninstall -y " + self.pkgName, shell=True)
        self.rmtree(os.path.join(self.path, "build"))
        self.rmtree(os.path.join(self.path, "dist"))
        self.rmtree(os.path.join(self.path, "static"))

    def run_build(self, kind):
        check_call([sys.executable, "-m", "build", "--no-isolation", f"--{kind}"], cwd=self.path)

    def check_correct_installation(self):
        # assert we can import buildbot_www
        # and that it has an endpoint with resource containing file "index.html"
        check_call([sys.executable, '-c', self.loadTestScript % dict(epName=self.epName)])

    def test_wheel(self):
        self.run_build("wheel")
        check_call("pip install build dist/*.whl", shell=True, cwd=self.path)
        self.check_correct_installation()

    def test_develop_via_pip(self):
        check_call("pip install build -e .", shell=True, cwd=self.path)
        self.check_correct_installation()

    def test_sdist(self):
        self.run_build("sdist")
        check_call("pip install build dist/*.tar.gz", shell=True, cwd=self.path)
        self.check_correct_installation()


class BuildbotConsolePkg(BuildbotWWWPkg):
    pkgName = "buildbot-console-view"
    pkgPaths = ["www", "console_view"]
    epName = "console_view"


class BuildbotWaterfallPkg(BuildbotWWWPkg):
    pkgName = "buildbot-waterfall-view"
    pkgPaths = ["www", "waterfall_view"]
    epName = "waterfall_view"



# This program is free software; you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Portions Copyright Buildbot Team Members
# Portions Copyright Marius Rieder <marius.rieder@durchmesser.ch>


from buildbot.process import logobserver


class WEObserver(logobserver.LogLineObserver):
    def __init__(self):
        super().__init__()
        self.warnings = []
        self.errors = []

    def outLineReceived(self, line):
        if line.startswith('W: '):
            self.warnings.append(line)
        elif line.startswith('E: '):
            self.errors.append(line)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from twisted.internet import defer

from buildbot import config
from buildbot.process import buildstep
from buildbot.process import logobserver
from buildbot.process.results import FAILURE
from buildbot.process.results import SUCCESS


class MaxQObserver(logobserver.LogLineObserver):
    def __init__(self):
        super().__init__()
        self.failures = 0

    def outLineReceived(self, line):
        if line.startswith('TEST FAILURE:'):
            self.failures += 1


class MaxQ(buildstep.ShellMixin, buildstep.BuildStep):
    flunkOnFailure = True
    name = "maxq"
    binary = 'run_maxq.py'

    failures = 0

    def __init__(self, testdir=None, **kwargs):
        if not testdir:
            config.error("please pass testdir")
        self.testdir = testdir

        kwargs = self.setupShellMixin(kwargs)
        super().__init__(**kwargs)
        self.observer = MaxQObserver()
        self.addLogObserver('stdio', self.observer)

    @defer.inlineCallbacks
    def run(self):
        command = [self.binary]
        command.append(self.testdir)

        cmd = yield self.makeRemoteShellCommand(command=command)
        yield self.runCommand(cmd)

        stdio_log = yield self.getLog('stdio')
        yield stdio_log.finish()

        self.failures = self.observer.failures

        # treat a nonzero exit status as a failure, if no other failures are
        # detected
        if not self.failures and cmd.didFail():
            self.failures = 1
        if self.failures:
            return FAILURE
        return SUCCESS

    def getResultSummary(self):
        if self.failures:
            return {'step': f"{self.failures} maxq failures"}
        return {'step': 'success'}



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Generator

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.db import sourcestamps
from buildbot.test import fakedb
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin
from buildbot.util import epoch2datetime

if TYPE_CHECKING:
    from buildbot.test.fakedb import FakeDBConnector

CREATED_AT = 927845299


def sourceStampKey(sourceStamp: sourcestamps.SourceStampModel):
    return (sourceStamp.repository, sourceStamp.branch, sourceStamp.created_at)


class Tests(TestReactorMixin, unittest.TestCase):
    db: FakeDBConnector

    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.master = yield fakemaster.make_master(self, wantDb=True)
        self.db = self.master.db

    @defer.inlineCallbacks
    def test_findSourceStampId_simple(self):
        self.reactor.advance(CREATED_AT)
        ssid = yield self.db.sourcestamps.findSourceStampId(
            branch='production',
            revision='abdef',
            repository='test://repo',
            codebase='cb',
            project='stamper',
        )
        ssdict = yield self.db.sourcestamps.getSourceStamp(ssid)
        self.assertEqual(
            ssdict,
            sourcestamps.SourceStampModel(
                branch='production',
                codebase='cb',
                patch=None,
                project='stamper',
                repository='test://repo',
                revision='abdef',
                ssid=ssid,
                created_at=epoch2datetime(CREATED_AT),
            ),
        )

    @defer.inlineCallbacks
    def test_findSourceStampId_simple_unique(self):
        ssid1 = yield self.db.sourcestamps.findSourceStampId(
            branch='production',
            revision='abdef',
            repository='test://repo',
            codebase='cb',
            project='stamper',
        )
        ssid2 = yield self.db.sourcestamps.findSourceStampId(
            branch='production',
            revision='xxxxx',  # different revision
            repository='test://repo',
            codebase='cb',
            project='stamper',
        )
        ssid3 = yield self.db.sourcestamps.findSourceStampId(  # same as ssid1
            branch='production',
            revision='abdef',
            repository='test://repo',
            codebase='cb',
            project='stamper',
        )
        self.assertEqual(ssid1, ssid3)
        self.assertNotEqual(ssid1, ssid2)

    @defer.inlineCallbacks
    def test_findSourceStampId_simple_unique_patch(self):
        ssid1 = yield self.db.sourcestamps.findSourceStampId(
            branch='production',
            revision='abdef',
            repository='test://repo',
            codebase='cb',
            project='stamper',
            patch_body=b'++ --',
            patch_level=1,
            patch_author='me',
            patch_comment='hi',
            patch_subdir='.',
        )
        ssid2 = yield self.db.sourcestamps.findSourceStampId(
            branch='production',
            revision='abdef',
            repository='test://repo',
            codebase='cb',
            project='stamper',
            patch_body=b'++ --',
            patch_level=1,
            patch_author='me',
            patch_comment='hi',
            patch_subdir='.',
        )
        # even with the same patch contents, we get different ids
        self.assertNotEqual(ssid1, ssid2)

    @defer.inlineCallbacks
    def test_findSourceStampId_patch(self):
        self.reactor.advance(CREATED_AT)
        ssid = yield self.db.sourcestamps.findSourceStampId(
            branch='production',
            revision='abdef',
            repository='test://repo',
            codebase='cb',
            project='stamper',
            patch_body=b'my patch',
            patch_level=3,
            patch_subdir='master/',
            patch_author='me',
            patch_comment="comment",
        )
        ssdict = yield self.db.sourcestamps.getSourceStamp(ssid)
        self.assertEqual(
            ssdict,
            sourcestamps.SourceStampModel(
                branch='production',
                codebase='cb',
                project='stamper',
                repository='test://repo',
                revision='abdef',
                created_at=epoch2datetime(CREATED_AT),
                ssid=ssid,
                patch=sourcestamps.PatchModel(
                    patchid=1,
                    author='me',
                    body=b'my patch',
                    comment='comment',
                    level=3,
                    subdir='master/',
                ),
            ),
        )

    @defer.inlineCallbacks
    def test_getSourceStamp_simple(self):
        yield self.db.insert_test_data([
            fakedb.SourceStamp(
                id=234,
                branch='br',
                revision='rv',
                repository='rep',
                codebase='cb',
                project='prj',
                created_at=CREATED_AT,
            ),
        ])
        ssdict = yield self.db.sourcestamps.getSourceStamp(234)

        self.assertEqual(
            ssdict,
            sourcestamps.SourceStampModel(
                ssid=234,
                created_at=epoch2datetime(CREATED_AT),
                branch='br',
                revision='rv',
                repository='rep',
                codebase='cb',
                project='prj',
                patch=None,
            ),
        )

    @defer.inlineCallbacks
    def test_getSourceStamp_simple_None(self):
        "check that NULL branch and revision are handled correctly"
        yield self.db.insert_test_data([
            fakedb.SourceStamp(
                id=234, branch=None, revision=None, repository='rep', codebase='cb', project='prj'
            ),
        ])
        ssdict = yield self.db.sourcestamps.getSourceStamp(234)

        self.assertIsInstance(ssdict, sourcestamps.SourceStampModel)
        self.assertEqual((ssdict.branch, ssdict.revision), (None, None))

    @defer.inlineCallbacks
    def test_getSourceStamp_patch(self) -> Generator[defer.Deferred, None, None]:
        yield self.db.insert_test_data([
            fakedb.Patch(
                id=99,
                patch_base64='aGVsbG8sIHdvcmxk',
                patch_author='bar',
                patch_comment='foo',
                subdir='/foo',
                patchlevel=3,
            ),
            fakedb.SourceStamp(id=234, patchid=99),
        ])
        res = yield self.db.sourcestamps.getSourceStamp(234)
        assert res is not None
        ssdict = res

        self.assertIsInstance(ssdict, sourcestamps.SourceStampModel)
        self.assertIsInstance(ssdict.patch, sourcestamps.PatchModel)
        self.assertEqual(ssdict.patch.body, b'hello, world')
        self.assertEqual(ssdict.patch.level, 3)
        self.assertEqual(ssdict.patch.author, 'bar')
        self.assertEqual(ssdict.patch.comment, 'foo')
        self.assertEqual(ssdict.patch.subdir, '/foo')

    @defer.inlineCallbacks
    def test_getSourceStamp_nosuch(self):
        ssdict = yield self.db.sourcestamps.getSourceStamp(234)

        self.assertEqual(ssdict, None)

    @defer.inlineCallbacks
    def test_getSourceStamps(self):
        yield self.db.insert_test_data([
            fakedb.Patch(
                id=99,
                patch_base64='aGVsbG8sIHdvcmxk',
                patch_author='bar',
                patch_comment='foo',
                subdir='/foo',
                patchlevel=3,
            ),
            fakedb.SourceStamp(
                id=234,
                revision='r',
                project='p',
                codebase='c',
                repository='rep',
                branch='b',
                patchid=99,
                created_at=CREATED_AT,
            ),
            fakedb.SourceStamp(
                id=235,
                revision='r2',
                project='p2',
                codebase='c2',
                repository='rep2',
                branch='b2',
                patchid=None,
                created_at=CREATED_AT + 10,
            ),
        ])
        db_sourcestamps = yield self.db.sourcestamps.getSourceStamps()

        self.assertEqual(
            sorted(db_sourcestamps, key=sourceStampKey),
            sorted(
                [
                    sourcestamps.SourceStampModel(
                        branch='b',
                        codebase='c',
                        project='p',
                        repository='rep',
                        revision='r',
                        created_at=epoch2datetime(CREATED_AT),
                        ssid=234,
                        patch=sourcestamps.PatchModel(
                            patchid=99,
                            author='bar',
                            body=b'hello, world',
                            comment='foo',
                            level=3,
                            subdir='/foo',
                        ),
                    ),
                    sourcestamps.SourceStampModel(
                        branch='b2',
                        codebase='c2',
                        project='p2',
                        repository='rep2',
                        revision='r2',
                        created_at=epoch2datetime(CREATED_AT + 10),
                        ssid=235,
                        patch=None,
                    ),
                ],
                key=sourceStampKey,
            ),
        )

    @defer.inlineCallbacks
    def test_getSourceStamps_empty(self):
        sourcestamps = yield self.db.sourcestamps.getSourceStamps()

        self.assertEqual(sourcestamps, [])

    @defer.inlineCallbacks
    def test_get_sourcestamps_for_buildset_one_codebase(self):
        yield self.db.insert_test_data([
            fakedb.Master(id=88, name="bar"),
            fakedb.Worker(id=13, name="one"),
            fakedb.Builder(id=77, name="A"),
            fakedb.SourceStamp(id=234, codebase="A", created_at=CREATED_AT, revision="aaa"),
            fakedb.Buildset(id=30, reason="foo", submitted_at=1300305712, results=-1),
            fakedb.BuildsetSourceStamp(sourcestampid=234, buildsetid=30),
        ])

        db_sourcestamps = yield self.db.sourcestamps.get_sourcestamps_for_buildset(30)

        expected = [
            sourcestamps.SourceStampModel(
                branch="master",
                codebase="A",
                created_at=epoch2datetime(CREATED_AT),
                patch=None,
                project="proj",
                repository="repo",
                revision="aaa",
                ssid=234,
            )
        ]

        self.assertEqual(
            sorted(db_sourcestamps, key=sourceStampKey), sorted(expected, key=sourceStampKey)
        )

    @defer.inlineCallbacks
    def test_get_sourcestamps_for_buildset_three_codebases(self):
        yield self.db.insert_test_data([
            fakedb.Master(id=88, name="bar"),
            fakedb.Worker(id=13, name="one"),
            fakedb.Builder(id=77, name="A"),
            fakedb.SourceStamp(id=234, codebase="A", created_at=CREATED_AT, revision="aaa"),
            fakedb.SourceStamp(id=235, codebase="B", created_at=CREATED_AT + 10, revision="bbb"),
            fakedb.SourceStamp(id=236, codebase="C", created_at=CREATED_AT + 20, revision="ccc"),
            fakedb.Buildset(id=30, reason="foo", submitted_at=1300305712, results=-1),
            fakedb.BuildsetSourceStamp(sourcestampid=234, buildsetid=30),
            fakedb.BuildsetSourceStamp(sourcestampid=235, buildsetid=30),
            fakedb.BuildsetSourceStamp(sourcestampid=236, buildsetid=30),
        ])

        db_sourcestamps = yield self.db.sourcestamps.get_sourcestamps_for_buildset(30)

        expected = [
            sourcestamps.SourceStampModel(
                branch="master",
                codebase="A",
                created_at=epoch2datetime(CREATED_AT),
                patch=None,
                project="proj",
                repository="repo",
                revision="aaa",
                ssid=234,
            ),
            sourcestamps.SourceStampModel(
                branch="master",
                codebase="B",
                created_at=epoch2datetime(CREATED_AT + 10),
                patch=None,
                project="proj",
                repository="repo",
                revision="bbb",
                ssid=235,
            ),
            sourcestamps.SourceStampModel(
                branch="master",
                codebase="C",
                created_at=epoch2datetime(CREATED_AT + 20),
                patch=None,
                project="proj",
                repository="repo",
                revision="ccc",
                ssid=236,
            ),
        ]

        self.assertEqual(
            sorted(db_sourcestamps, key=sourceStampKey), sorted(expected, key=sourceStampKey)
        )

    @defer.inlineCallbacks
    def do_test_getSourceStampsForBuild(self, rows, buildid, expected):
        yield self.db.insert_test_data(rows)

        sourcestamps = yield self.db.sourcestamps.getSourceStampsForBuild(buildid)

        self.assertEqual(
            sorted(sourcestamps, key=sourceStampKey), sorted(expected, key=sourceStampKey)
        )

    def test_getSourceStampsForBuild_OneCodeBase(self):
        rows = [
            fakedb.Master(id=88, name="bar"),
            fakedb.Worker(id=13, name='one'),
            fakedb.Builder(id=77, name='A'),
            fakedb.SourceStamp(id=234, codebase='A', created_at=CREATED_AT, revision="aaa"),
            # fakedb.Change(changeid=14, codebase='A', sourcestampid=234),
            fakedb.Buildset(id=30, reason='foo', submitted_at=1300305712, results=-1),
            fakedb.BuildsetSourceStamp(sourcestampid=234, buildsetid=30),
            fakedb.BuildRequest(
                id=19, buildsetid=30, builderid=77, priority=13, submitted_at=1300305712, results=-1
            ),
            fakedb.Build(
                id=50,
                buildrequestid=19,
                number=5,
                masterid=88,
                builderid=77,
                state_string="test",
                workerid=13,
                started_at=1304262222,
            ),
        ]

        expected = [
            sourcestamps.SourceStampModel(
                branch='master',
                codebase='A',
                created_at=epoch2datetime(CREATED_AT),
                patch=None,
                project='proj',
                repository='repo',
                revision='aaa',
                ssid=234,
            )
        ]

        return self.do_test_getSourceStampsForBuild(rows, 50, expected)

    def test_getSourceStampsForBuild_3CodeBases(self):
        rows = [
            fakedb.Master(id=88, name="bar"),
            fakedb.Worker(id=13, name='one'),
            fakedb.Builder(id=77, name='A'),
            fakedb.SourceStamp(id=234, codebase='A', created_at=CREATED_AT, revision="aaa"),
            fakedb.SourceStamp(id=235, codebase='B', created_at=CREATED_AT + 10, revision="bbb"),
            fakedb.SourceStamp(id=236, codebase='C', created_at=CREATED_AT + 20, revision="ccc"),
            # fakedb.Change(changeid=14, codebase='A', sourcestampid=234),
            fakedb.Buildset(id=30, reason='foo', submitted_at=1300305712, results=-1),
            fakedb.BuildsetSourceStamp(sourcestampid=234, buildsetid=30),
            fakedb.BuildsetSourceStamp(sourcestampid=235, buildsetid=30),
            fakedb.BuildsetSourceStamp(sourcestampid=236, buildsetid=30),
            fakedb.BuildRequest(
                id=19, buildsetid=30, builderid=77, priority=13, submitted_at=1300305712, results=-1
            ),
            fakedb.Build(
                id=50,
                buildrequestid=19,
                number=5,
                masterid=88,
                builderid=77,
                state_string="test",
                workerid=13,
                started_at=1304262222,
            ),
        ]

        expected = [
            sourcestamps.SourceStampModel(
                branch='master',
                codebase='A',
                created_at=epoch2datetime(CREATED_AT),
                patch=None,
                project='proj',
                repository='repo',
                revision='aaa',
                ssid=234,
            ),
            sourcestamps.SourceStampModel(
                branch='master',
                codebase='B',
                created_at=epoch2datetime(CREATED_AT + 10),
                patch=None,
                project='proj',
                repository='repo',
                revision='bbb',
                ssid=235,
            ),
            sourcestamps.SourceStampModel(
                branch='master',
                codebase='C',
                created_at=epoch2datetime(CREATED_AT + 20),
                patch=None,
                project='proj',
                repository='repo',
                revision='ccc',
                ssid=236,
            ),
        ]
        return self.do_test_getSourceStampsForBuild(rows, 50, expected)
